from __future__ import annotations

import copy
from dataclasses import dataclass


@dataclass
class MenuItem:
    """
    MenuItem data model.

    ## Attributes

    `name`: The name of the menu item.

    `description`: The description of the menu item.

    `price`: The price of the menu item.

    `category`: The category of the menu item.
    """

    name: str
    description: str
    price: float
    category: str

    @staticmethod
    def new_empty() -> MenuItem:
        """
        Creates a new empty MenuItem object.
        """
        return MenuItem(
            name="",
            description="",
            price=0.0,
            category="",
        )

    def copy(self) -> MenuItem:
        """
        Creates a copy of the MenuItem object.
        """
        return copy.copy(self)


# Initialize some data
MENU_ITEMS: list[MenuItem] = [
    MenuItem(
        name="Hamburger",
        description="A classic hamburger with lettuce, tomato, and onions",
        price=4.99,
        category="Burgers",
    ),
    MenuItem(
        name="Cheeseburger",
        description="A classic cheeseburger with lettuce, tomato, onions and cheese",
        price=5.99,
        category="Burgers",
    ),
    MenuItem(
        name="Fries",
        description="A side of crispy fries",
        price=2.99,
        category="Sides",
    ),
    MenuItem(
        name="Soda",
        description="A refreshing soda",
        price=1.99,
        category="Drinks",
    ),
    MenuItem(
        name="Salad",
        description="A fresh salad",
        price=4.99,
        category="Salads",
    ),
    MenuItem(
        name="Ice Cream",
        description="A sweet treat",
        price=3.99,
        category="Desserts",
    ),
]



from __future__ import annotations

# <additional-imports>
import functools

import rio

from .. import components as comps
from ..data_models import TodoAppSettings, TodoItem

# </additional-imports>


# <component>
@rio.page(
    name="Todo List",
    url_segment="",
)
class TodoListPage(rio.Component):
    """
    This page displays the entire GUI for the Todo App.

    There's a status message at the top, then a list of `TodoItem`s, and finally
    an input form that lets the user create a new `TodoItem`.
    """

    async def _add_new_todo_item(self, todo_item: TodoItem) -> None:
        # Append the new item to the list
        settings = self.session[TodoAppSettings]
        settings.todo_items.append(todo_item)

        # Save the settings and rebuild this component
        await self._on_settings_changed()

    async def _delete_todo_item(self, todo_item: TodoItem) -> None:
        # Remove the item from the list
        settings = self.session[TodoAppSettings]
        settings.todo_items.remove(todo_item)

        # Save the settings and rebuild this component
        await self._on_settings_changed()

    async def _on_settings_changed(self) -> None:
        # Re-attach the settings to save them
        self.session.attach(self.session[TodoAppSettings])

        # Rio doesn't know that this component needs to be rebuilt, since it
        # doesn't have any properties that have changed. We'll tell rio to
        # rebuild it by calling `self.force_refresh()`.
        await self.force_refresh()

    def build(self) -> rio.Component:
        settings = self.session[TodoAppSettings]

        # Display a status message depending on the number of unfinished items
        num_incomplete_items = sum(
            not todo_item.completed for todo_item in settings.todo_items
        )
        if num_incomplete_items > 0:
            status_text = f'You have {num_incomplete_items} unfinished task{"" if num_incomplete_items == 1 else "s"} 📋'
        elif settings.todo_items:
            status_text = "You have completed all your tasks 🎉"
        else:
            status_text = "You don't have anything to do 💤"

        return rio.Column(
            # Status message
            rio.Text(
                status_text,
                justify="center",
            ),
            # List of todo items
            rio.Column(
                *[
                    comps.TodoItemComponent(
                        todo_item,
                        on_completed=self._on_settings_changed,
                        on_deleted=functools.partial(
                            self._delete_todo_item, todo_item
                        ),
                    )
                    for todo_item in settings.todo_items
                ],
                rio.Spacer(),
                grow_y=True,
            ),
            # Input for new todo items
            comps.NewTodoItemInput(
                on_input=self._add_new_todo_item,
            ),
            margin=1,
            margin_bottom=0,
            spacing=1,
        )


# </component>



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY

from uniserde import JsonDoc

import rio.docs

from .fundamental_component import FundamentalComponent

__all__ = [
    "Popup",
]


@t.final
# @deprecations.component_kwarg_renamed(
#     since="0.9.2",
#     old_name="direction",
#     new_name="position",
# )
class Popup(FundamentalComponent):
    """
    A container which floats above other components.

    Popups are containers which float above the page when open. This allows you
    to keep your app clean by default, but present additional information or
    controls when needed.

    They take two children: The `anchor` is always visible and positions the
    popup. The `content` is located inside the popup and is only visible when
    the popup is open.

    The location popups appear at can be customized using the `direction`,
    `alignment` and `gap` attributes. Popups will do their best to honor those
    settings, but deviate if necessary to ensure they don't go off-screen.


    ## Attributes

    `anchor`: A component which is always visible and positions the popup.

    `content`: A component which is only visible when the popup is open.

    `color`: The color scheme to use for the popup's content. The popup will use the
        specified color as background, while content will automatically use one
        that is legible on top of it.

    `corner_radius`: The radius of the card's corners. If set to `None`, it
        is picked from the active theme.

    `position`: The location at which the popup opens, relative to the anchor.

    `alignment`: The alignment of the popup within the anchor. If the popup
        opens to the left or right, this is the vertical alignment, with `0`
        being the top and `1` being the bottom. If the popup opens to the top or
        bottom, this is the horizontal alignment, with `0` being the left and
        `1` being the right. Has no effect if the popup opens centered.

    `gap`: How much space to leave between the popup and the anchor. Has no
        effect popup opens centered. As all units in Rio, this is measured in
        font-heights.

    `is_open`: Whether the popup is currently open.


    ## Examples

    Most popups are opened by clicking a button. Here's an example that opens and
    closes a popup when a button is pressed:

    ```python
    class MyComponent(rio.Component):
        is_open: bool = False

        def on_button_press(self):
            self.is_open = not self.is_open

        def build(self) -> rio.Component:
            return rio.Popup(
                anchor=rio.Button(
                    "Open Popup",
                    on_press=self.on_button_press,
                ),
                content=rio.Card(
                    content=rio.Text(
                        "Hello World!",
                        justify="center",
                    ),
                    min_width=30,
                    min_height=10,
                ),
                # The popup will open above the anchor
                is_open=self.is_open,
                position="top",
            )
    ```
    """

    # TODO: Add example to docstring

    anchor: rio.Component
    content: rio.Component
    _: KW_ONLY
    color: rio.ColorSet | t.Literal["none"] = "hud"
    corner_radius: float | tuple[float, float, float, float] | None = None
    position: t.Literal[
        "auto", "left", "top", "right", "bottom", "center", "fullscreen"
    ] = "center"
    alignment: float = 0.5
    gap: float = 0.8
    is_open: bool = False

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "color": (
                "none"
                if self.color == "none"
                else self.session.theme._serialize_colorset(self.color)
            ),
            "corner_radius": (
                self.session.theme.corner_radius_medium
                if self.corner_radius is None
                else self.corner_radius
            ),
        }


Popup._unique_id_ = "Popup-builtin"



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY

import rio

from . import component

__all__ = [
    "Banner",
]

ICONS_AND_COLORS: dict[str, tuple[str, rio.ColorSet]] = {
    "info": ("material/info", "secondary"),
    "success": ("material/check_circle", "success"),
    "warning": ("material/warning", "warning"),
    "danger": ("material/error", "danger"),
}


@t.final
class Banner(component.Component):
    r"""
    Displays a short message to the user.

    Banners can either show a short text message to users, or disappear entirely
    if no message is set. Use them to inform the user about the result of an
    action, to give feedback on their input, or anything else that needs to be
    communicated.

    The messages have one of four styles: `"info"`, `"success"`, `"warning"`,
    and `"danger"`. These control the appearance of the notification bar, and
    allow you to quickly communicate the nature of the message to the user.


    ## Attributes

    `text`: The text to display. If `None` or empty, the banner will disappear
        entirely.

    `style`: Controls the appearance of the banner. The style is one of
        `"info"`, `"success"`, `"warning"` and `"danger"`. Depending on the
        value the banner may change its colors and icon.

    `markdown`: If `True`, the banner text will be rendered as Markdown,
        otherwise it will be rendered as plain text.

    ## Examples

    This shows the most simple usage of a `Banner`:

    ```python
    rio.Banner(
        text="This is a banner",
        style="info",
    )
    ```

    `Banner`s are commonly used to inform the users of about the result of an
    action. You can easily achieve this by adding a banner to a column and
    setting its text when that action completes (or fails!). Here's an example:

    ```python
    class MyComponent(rio.Component):
        banner_text: str = ""

        def on_press_button(self) -> None:
            self.banner_text = "Button pressed!"

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Banner(
                    text=self.banner_text,
                    style="info",
                ),
                rio.Button(
                    content="Press me!",
                    on_press=self.on_press_button,
                ),
                spacing=1,
            )
    ```
    """

    text: str | None
    style: t.Literal["info", "success", "warning", "danger"]

    _: KW_ONLY
    markdown: bool = False
    icon: str | None = None

    def build(self) -> rio.Component:
        # Early out: Nothing to show
        if self.text is None:
            return rio.Spacer(grow_x=False, grow_y=False)

        text = self.text.strip()
        if not text:
            return rio.Spacer(grow_x=False, grow_y=False)

        # Prepare the style
        try:
            icon, style_name = ICONS_AND_COLORS[self.style]
        except KeyError:
            raise ValueError(f"Invalid style: {self.style}")

        if self.icon is not None:
            icon = self.icon

        # Prepare the text child
        if self.markdown:
            text_child = rio.Markdown(
                text,
                grow_x=True,
            )
        else:
            text_child = rio.Text(
                text,
                grow_x=True,
                overflow="wrap",
                justify="left",
            )

        # Build the result
        return rio.Card(
            content=rio.Row(
                rio.Icon(icon),
                text_child,
                spacing=0.5,
                margin=0.5,
            ),
            color=style_name,  # type: ignore
            corner_radius=self.session.theme.corner_radius_small,
            elevate_on_hover=False,
        )



import typing as t

from .fundamental_component import FundamentalComponent

__all__ = ["Html"]


@t.final
class Html(FundamentalComponent):
    """
    Displays raw HTML.

    The `Html` component allows you to embed arbitrary HTML in your app. It
    takes HTML code as input and renders it.

    To embed an entire website, use the `Website` component instead.


    ## Attributes

    `html`: The HTML to render.


    ## Examples

    This minimal example will render a simple HTML code:

    ```python
    rio.Html('''
    <h1>Hello World</h1>
    <p>Welcome to rio!</p>
    ''')
    """

    html: str


Html._unique_id_ = "Html-builtin"



from __future__ import annotations

import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "ProgressBar",
]


@t.final
class ProgressBar(FundamentalComponent):
    """
    A progress indicator in the shape of a horizontal bar.

    `ProgressBar` conveys to the user that activity is ongoing. It can either
    display the exact progress as a fraction from 0 to 1, or it can display an
    indeterminate progress animation, which is useful when the exact progress
    isn't known.

    The circular counterpart to this component is the `ProgressCircle`.


    ## Attributes

    `progress`: The progress to display, as a fraction from 0 to 1. If `None`,
        the progress indicator will be indeterminate.

    `color`: The color scheme of the progress indicator. Keeping the default
        is recommended, but it may make sense to change the color in
        case the default is hard to perceive on your background.

    `rounded`: Whether the corners of the progress bar should be rounded.

    ## Examples

    Here's a minimal example displaying a progress circle that is 40% complete:

    ```python
    rio.ProgressBar(progress=0.4)
    ```

    You don't have to pass any progress to `rio.ProgressBar`. If you don't,
    it will indicate to the user that something is happening, without showing
    any specific progress.

    ```python
    rio.ProgressBar()
    ```
    """

    progress: float | None
    color: rio.ColorSet
    rounded: bool

    def __init__(
        self,
        progress: float | None = None,
        *,
        color: rio.ColorSet = "keep",
        rounded: bool = True,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        """
        ## Parameters

        `progress`: The progress to display, as a fraction from 0 to 1. If `None`,
            the progress indicator will be indeterminate.

        `color`: The color scheme of the progress indicator. Keeping the default
            is recommended, but it may make sense to change the color in case
            the default is hard to perceive on your background.


        `rounded`: Whether the corners of the progress bar should be rounded.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.progress = progress
        self.color = color
        self.rounded = rounded


ProgressBar._unique_id_ = "ProgressBar-builtin"



from __future__ import annotations

import typing as t
from dataclasses import dataclass

# Never import * from typing_extensions! It breaks `Any` on 3.10, preventing
# users from connecting. Ask me how I know.
import typing_extensions as te
import uniserde

import rio

__all__ = ["BuildData", "ComponentLayout", "InitialClientMessage"]


@dataclass
class BuildData:
    build_result: rio.Component

    all_children_in_build_boundary: set[rio.Component]

    # Keep track of how often this component has been built. This is used by
    # components to determine whether they are still part of their parent's current
    # build output.
    build_generation: int


@dataclass
class InitialClientMessage(uniserde.Serde):
    # The URL the client used to connect to the website. This can be quite
    # different from the URLs we see in FastAPI requests, because proxies like
    # nginx can alter it. For example, the client may be connecting via https
    # while rio only sees http.
    url: str

    # Don't annotate this as JsonDoc because uniserde doesn't support unions
    user_settings: dict[str, t.Any]

    prefers_light_theme: bool

    # List of RFC 5646 language codes. Most preferred first. May be empty!
    preferred_languages: list[str]

    # The names for all months, starting with January
    month_names_long: tuple[
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
        str,
    ]

    # The names of all weekdays, starting with monday
    day_names_long: tuple[
        str,
        str,
        str,
        str,
        str,
        str,
        str,
    ]

    # The format string for dates, as used by Python's `strftime`
    date_format_string: str

    # IANA timezone
    timezone: str

    # Separators for number rendering
    decimal_separator: str
    thousands_separator: str

    # Window Information
    window_width: float
    window_height: float

    @classmethod
    def from_defaults(
        cls,
        *,
        url: str,
        user_settings: uniserde.JsonDoc = {},
    ) -> te.Self:
        """
        Convenience method for creating default settings when they don't really
        matter: unit tests, crawlers, etc.
        """
        return cls(
            url=url,
            user_settings=user_settings,
            prefers_light_theme=True,
            preferred_languages=["en-US"],
            month_names_long=(
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ),
            day_names_long=(
                "Monday",
                "Tuesday",
                "Wednesday",
                "Thursday",
                "Friday",
                "Saturday",
                "Sunday",
            ),
            date_format_string="%Y-%m-%d",
            timezone="America/New_York",
            decimal_separator=".",
            thousands_separator=",",
            window_width=1920,
            window_height=1080,
        )


# class CliConfig(uniserde.Config):
#     # Whether to automatically open the browser when running an app using `rio
#     # run`. This has no effect on functions that are explicitly meant to open
#     # browsers, such as `app.run_in_browser`.
#     open_browser_on_startup: bool = True


@dataclass
class ComponentLayout(uniserde.Serde):
    # The minimum amount of size needed by the component. The width is
    # calculated first, meaning the height can depend on the width. (i.e. a
    # text's height depends on the width because it wraps)
    natural_width: float
    natural_height: float

    # Components can request more space than their natural size if a size was
    # explicitly provided on the Python-side. This value is the maximum of the
    # natural size and any explicitly provided size.
    requested_inner_width: float
    requested_inner_height: float

    # The requested width after scrolling, alignment and margin.
    requested_outer_width: float
    requested_outer_height: float

    # The amount of space allocated to the component before scrolling,
    # alignment and margin.
    allocated_outer_width: float
    allocated_outer_height: float

    # The amount of space allocated to the component after scrolling,
    # alignment and margin.
    allocated_inner_width: float
    allocated_inner_height: float

    # The component's position relative to the viewport before scrolling,
    # alignment and margin.
    left_in_viewport_outer: float
    top_in_viewport_outer: float

    # The component's position relative to the viewport after scrolling,
    # alignment and margin.
    left_in_viewport_inner: float
    top_in_viewport_inner: float

    # The id of the parent component, unless this is the root component.
    parent_id: int | None


@dataclass
class UnittestComponentLayout(ComponentLayout):
    # Additional, component-specific information
    aux: dict[str, t.Any]


@dataclass
class UnittestClientLayoutInfo:
    window_width: float
    window_height: float

    component_layouts: dict[int, UnittestComponentLayout]



from __future__ import annotations

import hashlib
import importlib.util
import mimetypes
import os
import re
import secrets
import socket
import sys
import typing as t
from dataclasses import dataclass
from io import BytesIO, StringIO
from pathlib import Path

import imy.assets
import typing_extensions as te
from PIL.Image import Image
from yarl import URL

import rio

__all__ = [
    "EventHandler",
    "FileInfo",
    "ImageLike",
    "escape_markdown",
    "escape_markdown_code",
    "i_know_what_im_doing",
]


_SECURE_HASH_SEED: bytes = secrets.token_bytes(32)

# Expose common paths on the filesystem
PACKAGE_ROOT_DIR = Path(__file__).absolute().parent
PROJECT_ROOT_DIR = PACKAGE_ROOT_DIR.parent

RIO_ASSETS_DIR = PACKAGE_ROOT_DIR / "assets"
HOSTED_ASSETS_DIR = RIO_ASSETS_DIR / "hosted"

FRONTEND_FILES_DIR = PACKAGE_ROOT_DIR / "frontend files"
FRONTEND_ASSETS_DIR = FRONTEND_FILES_DIR / "assets"

RIO_LOGO_ASSET_PATH = HOSTED_ASSETS_DIR / "rio_logos/rio_logo_square.png"

SNIPPETS_DIR = PACKAGE_ROOT_DIR / "snippets" / "snippet-files"

if os.name == "nt":
    USER_CACHE_DIR = Path.home() / "AppData" / "Local" / "Cache"
    USER_CONFIG_DIR = Path.home() / "AppData" / "Roaming"
else:
    USER_CACHE_DIR = Path.home() / ".cache"
    USER_CONFIG_DIR = Path.home() / ".config"


# Constants & types
_READONLY = object()
T = t.TypeVar("T")
Readonly = te.Annotated[T, _READONLY]

ImageLike = Path | Image | URL | bytes


ASSET_MANAGER: imy.assets.AssetManager = imy.assets.AssetManager(
    xz_dir=RIO_ASSETS_DIR,
    cache_dir=USER_CACHE_DIR / "rio",
    version=rio.__version__,
)


# Precompiled regexes
MARKDOWN_ESCAPE = re.compile(r"([\\`\*_\{\}\[\]\(\)#\+\-.!])")
MARKDOWN_CODE_ESCAPE = re.compile(r"([\\`])")


I_KNOW_WHAT_IM_DOING = set[object]()


def i_know_what_im_doing(thing: t.Callable):
    I_KNOW_WHAT_IM_DOING.add(thing)
    return thing


def secure_string_hash(*values: str, hash_length: int = 32) -> str:
    """
    Returns a reproducible, secure hash for the given values.

    The order of values matters. In addition to the given values a global seed
    is added. This seed is generated once when the module is loaded, meaning the
    result is not suitable to be persisted across sessions.
    """

    hasher = hashlib.blake2b(
        _SECURE_HASH_SEED,
        digest_size=hash_length,
    )

    for value in values:
        hasher.update(value.encode("utf-8"))

    return hasher.hexdigest()


@dataclass(frozen=True)
class FileInfo:
    """
    Contains information about a file.

    When asking the user to select a file, this class is used to represent the
    file. It contains metadata about the file, and can also be used to access
    the file's contents.

    Be careful when running your app as a webserver, since files will need to be
    uploaded by the user, which is a potentially very slow operation.


    ## Attributes

    `name`: The name of the file, including the extension.

    `size_in_bytes`: The size of the file, in bytes.

    `media_type`: The MIME type of the file, for example `text/plain` or
        `image/png`.
    """

    name: str
    size_in_bytes: int
    media_type: str
    _contents: bytes

    async def read_bytes(self) -> bytes:
        """
        Asynchronously reads the entire file as `bytes`.

        Reads and returns the entire file as a `bytes` object. If you know that
        the file is text, consider using `read_text` instead.
        """
        return self._contents

    async def read_text(self, *, encoding: str = "utf-8") -> str:
        """
        Asynchronously reads the entire file as text.

        Reads and returns the entire file as a `str` object. The file is decoded
        using the given `encoding`. If you don't know that the file is valid
        text, use `read_bytes` instead.

        ## Parameters

        encoding: The encoding to use when decoding the file.

        ## Raises

        UnicodeDecodeError: The file could not be decoded using the given
            `encoding`.
        """
        return self._contents.decode(encoding)

    @t.overload
    async def open(self, type: t.Literal["r"]) -> StringIO: ...

    @t.overload
    async def open(self, type: t.Literal["rb"]) -> BytesIO: ...

    async def open(
        self, type: t.Literal["r", "rb"] = "r"
    ) -> StringIO | BytesIO:
        """
        Asynchronously opens the file, as though it were a regular file on this
        device.

        Opens and returns the file as a file-like object. If 'r' is specified,
        the file is opened as text. If 'rb' is specified, the file is opened as
        bytes.

        Returns a file-like object containing the file's contents.

        ## Parameters

        type: The mode to open the file in. 'r' for text, 'rb' for bytes.
        """
        # Bytes
        if type == "rb":
            return BytesIO(await self.read_bytes())

        # UTF
        if type == "r":
            return StringIO(await self.read_text())

        # Invalid
        raise ValueError("Invalid type. Expected 'r' or 'rb'.")


T = t.TypeVar("T")
P = t.ParamSpec("P")

EventHandler = t.Callable[P, t.Any | t.Awaitable[t.Any]] | None


def make_url_relative(base: URL, other: URL) -> URL:
    """
    Returns `other` as a relative URL to `base`. Raises a `ValueError` if
    `other` is not a child of `base`.

    This will never generate URLs containing `..`. If those would be needed a
    `ValueError` is raised.
    """
    # Verify the URLs have the same scheme and host
    if base.scheme != other.scheme:
        raise ValueError(
            f'URLs have different schemes: "{base.scheme}" and "{other.scheme}"'
        )

    if base.host != other.host:
        raise ValueError(
            f'URLs have different hosts: "{base.host}" and "{other.host}"'
        )

    # Get the path segments of the URLs
    base_parts = base.parts
    other_parts = other.parts

    # Strip empty segments
    if base_parts and base_parts[-1] == "":
        base_parts = base_parts[:-1]

    if other_parts and other_parts[-1] == "":
        other_parts = other_parts[:-1]

    # See if the base is a parent of the other URL
    if base_parts != other_parts[: len(base_parts)]:
        raise ValueError(f'"{other}" is not a child of "{base}"')

    # Remove the common parts from the URL
    other_parts = other_parts[len(base_parts) :]
    return URL.build(
        path="/".join(other_parts),
        query=other.query,
        fragment=other.fragment,
    )


def escape_markdown(text: str) -> str:
    """
    Escape text such that it appears as-is when rendered as markdown.

    Given any text, this function returns a string which, when rendered as
    markdown, will look identical to the original text.


    ## Parameters

    `text`: The text to escape.
    """
    # TODO: Find a proper function for this. The current one is a total hack.
    return re.sub(MARKDOWN_ESCAPE, r"\\\1", text)


def escape_markdown_code(text: str) -> str:
    """
    Escape text such that it appears as-is inside a markdown code block.

    Given any text, this function returns a string which, when rendered inside a
    markdown code block, will look identical to the original text.


    ## Parameters

    `text`: The text to escape.
    """
    # TODO: Find a proper function for this. The current one is a total hack.
    return MARKDOWN_CODE_ESCAPE.sub(r"\\\1", text)


def choose_free_port(host: str) -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind((host, 0))
        return sock.getsockname()[1]


def port_is_free(host: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        try:
            sock.bind((host, port))
            return True
        except OSError as err:
            return False


def ensure_valid_port(host: str, port: int | None) -> int:
    if port is None:
        return choose_free_port(host)

    return port


def first_non_null(*values: T | None) -> T:
    """
    Returns the first non-`None` value, or raises a `ValueError` if all values
    are `None`.
    """

    for value in values:
        if value is not None:
            return value

    raise ValueError("At least one value must be non-`None`")


def _repr_build_function(build_function: t.Callable[[], rio.Component]) -> str:
    """
    Return a recognizable name for the provided function such as
    `Component.build`.
    """

    try:
        self = build_function.__self__  # type: ignore
    except AttributeError:
        try:
            return build_function.__qualname__
        except AttributeError:
            return repr(build_function)

    return f"{type(self).__name__}.{build_function.__name__}"


def safe_build(build_function: t.Callable[[], rio.Component]) -> rio.Component:
    """
    Calls a build function and returns its result. This differs from just
    calling the function directly, because it catches any exceptions and returns
    a placeholder component instead. It will also ensure the returned value is
    in fact a component.
    """

    # Build the component
    try:
        build_result = build_function()

    # The function has crashed. Return a placeholder instead
    except Exception as err:
        build_function_repr = _repr_build_function(build_function)

        rio._logger.exception(
            f"An exception occurred in `{build_function_repr}`"
        )

        # Screw circular imports
        from rio.components.error_placeholder import ErrorPlaceholder

        placeholder_component = ErrorPlaceholder(
            f"`{build_function_repr}` has crashed", repr(err)
        )
    else:
        # Make sure the result meets expectations
        if isinstance(build_result, rio.Component):  # type: ignore (unnecessary isinstance)
            # All is well
            return build_result

        build_function_repr = _repr_build_function(build_function)

        rio._logger.error(
            f"The output of `build` methods must be instances of"
            f" `rio.Component`, but `{build_function_repr}` returned `{build_result!r}`"
        )

        # Screw circular imports
        from rio.components.error_placeholder import ErrorPlaceholder

        placeholder_component = ErrorPlaceholder(
            f"`{build_function_repr}` has returned an invalid result",
            f"Build functions must return instances of `rio.Component`, but the result was {build_result!r}",
        )

    # Save the error in the session, for testing purposes
    placeholder_component.session._crashed_build_functions[build_function] = (
        placeholder_component.error_details
    )

    return placeholder_component


def normalize_url(url: rio.URL) -> rio.URL:
    """
    Returns a normalized version of the given URL.

    This returns a new URL instance which is identical to the given URL, but
    with the following guarantees:

    - The URL is lowercase
    - The URL has no trailing slashes
    """
    path = url.path.rstrip("/")
    path = path.lower()
    return url.with_path(path)


def load_module_from_path(file_path: Path, *, module_name: str | None = None):
    if module_name is None:
        module_name = file_path.stem

    try:
        module = sys.modules[module_name]
    except KeyError:
        pass
    else:
        if module.__file__ == str(file_path.absolute()):
            return module

        raise ImportError(
            f"The file {file_path} cannot be imported because a module named"
            f" {module_name!r} already exists."
        )

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)  # type: ignore (wtf?)

    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore (wtf?)

    return module


def is_python_script(path: Path) -> bool:
    return path.suffix in (".py", ".pyc", ".pyd", ".pyo", ".pyw")


def normalize_file_type(file_type: str) -> str:
    """
    Converts different file type formats into a common one.

    This function takes various formats of file types, such as file extensions
    (e.g., ".pdf", "PDF", "*.pdf") or MIME types (e.g., "application/pdf"), and
    converts them into a standardized file extension. The result is always
    lowercase and without any leading dots or wildcard characters.

    This is best-effort. If the input type is invalid or unknown, the cleaned
    input may not be accurate.

    Examples:
        >>> standardize_file_type("pdf")
        'pdf'
        >>> standardize_file_type(".PDF")
        'pdf'
        >>> standardize_file_type("*.pdf")
        'pdf'
        >>> standardize_file_type("application/pdf")
        'pdf'
    """
    # Normalize the input string
    file_type = file_type.lower().strip()

    # If this is a MIME type, guess the extension
    if "/" in file_type:
        guessed_type = mimetypes.guess_extension(file_type, strict=False)

        if guessed_type is None:
            file_type = file_type.rsplit("/", 1)[-1]
        else:
            file_type = guessed_type.lstrip(".")

    # If it isn't a MIME type, convert it to one anyway. Some file types have
    # multiple commonly used extensions. This will always map them to the same
    # one. For example "jpeg" and "jpg" are both mapped to "jpg".
    else:
        guessed_type, _ = mimetypes.guess_type(
            f"file.{file_type}", strict=False
        )

        if guessed_type is None:
            file_type = file_type.lstrip(".*")
        else:
            guessed_type = mimetypes.guess_extension(
                guessed_type,
                strict=False,
            )

            assert guessed_type is not None
            file_type = guessed_type.lstrip(".")

    # Done
    return file_type


def soft_sort(elements: list[T], key: t.Callable[[T], int | None]) -> None:
    """
    Sorts the given list in-place, allowing for `None` values in the key.

    This function tries to match sorting that humans expect when they assign
    positions to only some items. For example, an item with key `0` or `1`
    should be at the start. Likewise, an item with `9999` should be towards the
    end of the list.

    This function places elements with a key at the specified locations. It then
    assigns any leftover items to the remaining positions in the list, in the
    same order they were before sorting.

    Note that there is one special case where this may not work as intended: If
    two items have the same key assigned they obviously cannot be placed at the
    same position. In that case the original order between them is preserved,
    and followup items shifted as needed.
    """

    # To ensure we can keep the original order, add the current index to all
    # items.
    keyed_elements = [
        (element, index, key(element)) for index, element in enumerate(elements)
    ]

    # Assign positions to all elements with a provided key. This maps the
    # original index to the assigned position.
    assigned_positions: dict[int, int] = {}

    for _, original_index, key_value in keyed_elements:
        if key_value is None:
            continue

        assigned_positions[original_index] = key_value

    # Assign positions to all other elements
    ii = 0

    for _, original_index, key_value in keyed_elements:
        if key_value is not None:
            continue

        while ii in assigned_positions.values():
            ii += 1

        assigned_positions[original_index] = ii
        ii += 1

    # Sort the elements, taking care to preserve the original order between
    # items with the same key.
    def regular_sort_key(item: tuple[T, int, int | None]) -> tuple[int, int]:
        _, original_index, _ = item
        return assigned_positions[original_index], original_index

    keyed_elements.sort(key=regular_sort_key)

    # Reassign the elements to the original list
    elements.clear()

    for element, _, _ in keyed_elements:
        elements.append(element)



from __future__ import annotations

import typing as t
from dataclasses import field

import typing_extensions as te

import rio

from .component import Component

__all__ = ["LabeledColumn"]


@t.final
class LabeledColumn(Component):
    """
    A container that lays out its children in a column, with labels for each
    child.

    `LabeledColumn` is a container that lays out its children in a column, with
    labels for each child. It is similar to `Column`, but allows you to specify
    labels for each child.


    ## Example

    This minimal example will show a container with a column layout and labels:

    ```python
    rio.LabeledColumn(
        {
            "First Name": rio.TextInput(),
            "Last Name": rio.TextInput(),
            "Age": rio.TextInput(),
        }
    )
    ```

    ## Metadata

    `public`: False
    """

    _child_list: list[Component] = field(init=False)

    def __init__(
        self,
        content: t.Mapping[str, rio.Component],
        *,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.content = content

    @property
    def content(self) -> t.Mapping[str, Component]:
        return self._content

    @content.setter
    def content(self, children: t.Mapping[str, Component]) -> None:
        self._content = dict(children)
        self._child_list = list(children.values())

    def add(self, label: str, child: rio.Component) -> te.Self:
        """
        Appends a child component.

        Appends a child component to the end and then returns the
        `LabeledColumn`, which makes method chaining possible.
        """
        self._content[label] = child
        self._child_list.append(child)

        return self

    def build(self) -> Component:
        rows = [
            [
                rio.Text(label, justify="right", selectable=False),
                child,
            ]
            for label, child in self.content.items()
        ]

        return rio.Grid(*rows, row_spacing=0.1, column_spacing=0.2)



import time
import typing as t
from pathlib import Path

import watchfiles

from ... import project_config
from . import run_models


class FileWatcherWorker:
    def __init__(
        self,
        *,
        push_event: t.Callable[[run_models.Event], None],
        proj: project_config.RioProjectConfig,
    ) -> None:
        self.push_event = push_event
        self.proj = proj

    async def run(self) -> None:
        """
        Watch the project directory for changes and report them as events.
        """
        # Watch all files
        filter = watchfiles.DefaultFilter()

        # Watch the project directory
        async for changes in watchfiles.awatch(
            self.proj.project_directory, watch_filter=filter
        ):
            for change, path in changes:
                path = Path(path)

                # Not all files trigger a reload
                if not self.proj.file_is_path_of_project(path):
                    continue

                # Report the change
                self.push_event(
                    run_models.FileChanged(
                        time.monotonic_ns(),
                        path,
                    )
                )



from __future__ import annotations

import typing as t

import typing_extensions as te
from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Row",
    "Column",
]


class _LinearContainer(FundamentalComponent):
    children: list[rio.Component]
    spacing: float = 0.0
    proportions: t.Literal["homogeneous"] | t.Sequence[float] | None = None

    # Don't let @dataclass generate a constructor
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add(self, child: rio.Component) -> te.Self:
        self.children.append(child)
        return self

    def _custom_serialize_(self) -> JsonDoc:
        return {"proportions": self.proportions}  # type: ignore (variance)


@t.final
class Row(_LinearContainer):
    """
    A container that lays out its children horizontally.

    `Row`s are one of the most common components in Rio. They take any number of
    children and lay them out horizontally, with the first one on the left, the
    second one to its right, and so on. All components in `Row`s occupy the full
    height of their parent.

    The `Row`'s horizontal counterpart is the `Column`. A similar component, but
    stacking its children in the Z direction, is the `Stack`.


    ## Allocation of extra space

    When a `Row` has more horizontal space available than it needs, it will
    evenly distribute the extra space among all child components whose `grow_x`
    attribute is `True`.

    If no child is set to grow horizontally, the extra space is evenly
    distributed among all children. This is why components in a `Row` can
    sometimes become unexpectedly large. If you don't want that to happen, you
    can either tell Rio which children should receive the extra space by setting
    their `grow_x` to `True`, or you can set the `Row`s `align_x` to something
    other than `None`, which will cause the `Row` to only take up as much space
    as necessary and position itself in the available space.

    For more details, see the [layouting
    quickstart](https://rio.dev/docs/howto/layout-guide).


    ## Proportions

    Sometimes you want the widths of the children to be in some sort of
    relation. For example, you may want two children to have the same width.
    This can be achieved with the `proportions` parameter. Passing
    `proportions=[1, 1]` will make both children have the same width,
    `proportions=[1, 2]` would make the 2nd child twice the width of the 1st
    child, and so on.

    As a shortcut, you can also pass `proportions="homogeneous"` to make all
    children the same width.


    ## Attributes

    `children`: The components to place in this `Row`.

    `spacing`: How much empty space to leave between two adjacent children. No
            spacing is added before the first child or after the last child.

    `proportions`: If set, the children will grow according to these
        proportions.

        - `homogeneous`: All children will grow equally.
        - A list of floats: Each child will grow according to its proportion.
        - `None`: Extra space will be evenly distributed among children with
            `width='grow'`.


    ## Examples

    This minimal example will display a `Row` with two text components:

    ```python
    rio.Row(rio.Text("Hello"), rio.Text("World!"))
    ```

    `Row`s are used to line up multiple components horizontally. This example
    arranges a `rio.Icon` and two `rio.Text` components in a row and neatly
    wraps them in a Card.

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Card(
                content=rio.Row(
                    rio.Text("Hello"),
                    rio.Icon(icon="material/star"),
                    rio.Text("World!"),
                    spacing=1,
                    margin=1,
                ),
            )
    ```
    """

    def __init__(
        self,
        *children: rio.Component,
        spacing: float = 0.0,
        proportions: t.Literal["homogeneous"] | t.Sequence[float] | None = None,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.children = list(children)
        self.spacing = spacing
        self.proportions = proportions

    def add(self, child: rio.Component) -> te.Self:
        """
        Appends a child component.

        Appends a child component to the end of the row and returns the `Row`.
        This means you can chain multiple `add` calls:

        ```python
        rio.Row().add(child1).add(child2)
        ```

        ## Parameters

        `child`: The child component to append.
        """
        return super().add(child)


Row._unique_id_ = "Row-builtin"


@t.final
class Column(_LinearContainer):
    """
    A container that lays out its children vertically.

    `Column`s are one of the most common components in Rio. They take any number
    of children and lay them out vertically, with the first one at the top, the
    second one below that, and so on. All components in `Column`s occupy the
    full width of their parent.

    The `Column`'s horizontal counterpart is the `Row`. A similar component, but
    stacking its children in the Z direction, is the `Stack`.


    ## Allocation of extra space

    When a `Column` has more vertical space available than it needs, it will
    evenly distribute the extra space among all child components whose `grow_y`
    attribute is `True`.

    If no child is set to grow vertically, the extra space is evenly distributed
    among all children. This is why components in a `Column` can sometimes
    become unexpectedly large. If you don't want that to happen, you can either
    tell rio which children should receive the extra space by setting their
    `grow_y` to `True`, or you can set the `Column`s `align_y` to something
    other than `None`, which will cause the `Column` to only take up as much
    space as necessary and position itself in the available space.

    For more details, see the [layouting
    quickstart](https://rio.dev/docs/howto/layout-guide).


    ## Proportions

    Sometimes you want the heights of the children to be in some sort of
    relation. For example, you may want two children to have the same height.
    This can be achieved with the `proportions` parameter. Passing
    `proportions=[1, 1]` will make both children have the same height,
    `proportions=[1, 2]` would make the 2nd child twice the height of the 1st
    child, and so on.

    As a shortcut, you can also pass `proportions="homogeneous"` to make all
    children the same height.


    ## Attributes

    `children`: The components to place in this `Column`.

    `spacing`: How much empty space to leave between two adjacent children. No
        spacing is added before the first child or after the last child.

    `proportions`: If set, the children will grow according to these proportions.
        - `homogeneous`: All children will grow equally.
        - A list of floats: Each child will grow according to its proportion.
        - `None`: Extra space will be evenly distributed among children with
            `height='grow'`.


    ## Examples

    This minimal example will display a `Column` with two text components:

    ```python
    rio.Column(rio.Text("Hello"), rio.Text("World!"))
    ```

    `Columns`s are used to line up multiple components vertically. This example
    arranges a `rio.Icon` and two `rio.Text` components in a column and neatly
    wraps them in a Card.

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Card(
                content=rio.Column(
                    rio.Text("Hello"),
                    rio.Icon("material/castle"),
                    rio.Text("World!"),
                    spacing=1,
                    margin=1,
                ),
            )
    ```
    """

    def __init__(
        self,
        *children: rio.Component,
        spacing: float = 0.0,
        proportions: t.Literal["homogeneous"] | t.Sequence[float] | None = None,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.children = list(children)
        self.spacing = spacing
        self.proportions = proportions

    def add(self, child: rio.Component) -> te.Self:
        """
        Appends a child component.

        Appends a child component to the end of the column and returns the
        `Column`. This means you can chain multiple `add` calls:

        ```python
        rio.Column().add(child1).add(child2)
        ```

        ## Parameters

        `child`: The child component to append.
        """
        return super().add(child)


Column._unique_id_ = "Column-builtin"



from __future__ import annotations

import rio


# <component>
class Navbar(rio.Component):
    """
    A navbar with a fixed position and responsive width.
    """

    # Make sure the navbar will be rebuilt when the app navigates to a different
    # page. While Rio automatically detects state changes and rebuilds
    # components as needed, navigating to other pages is not considered a state
    # change, since it's not stored in the component.
    #
    # Instead, we can use Rio's `on_page_change` event to trigger a rebuild of
    # the navbar when the page changes.
    @rio.event.on_page_change
    async def on_page_change(self) -> None:
        # Rio comes with a function specifically for this. Whenever Rio is
        # unable to detect a change automatically, use this function to force a
        # refresh.
        await self.force_refresh()

    def build(self) -> rio.Component:
        # Which page is currently active? This will be used to highlight the
        # correct navigation button.
        #
        # `active_page_instances` contains `rio.ComponentPage` instances that
        # that are created during app creation. Since multiple pages can be
        # active at a time (e.g. /foo/bar/baz), this is a list rather than just
        # a single page.
        active_page = self.session.active_page_instances[0]
        active_page_url_segment = active_page.url_segment

        # The navbar should appear above all other components. This is easily
        # done by using a `rio.Overlay` component.
        return rio.Overlay(
            # Use a rectangle for visual separation
            rio.Rectangle(
                content=rio.Row(
                    # Links can be used to navigate to other pages and
                    # external URLs. You can pass either a simple string, or
                    # another component as their content.
                    rio.Link(
                        rio.IconButton(
                            "rio/logo",
                            style="plain-text",
                            min_size=2.5,
                        ),
                        "/",
                    ),
                    # This spacer will take up any superfluous space,
                    # effectively pushing the subsequent buttons to the
                    # right.
                    rio.Spacer(),
                    # By sticking buttons into a `rio.Link`, we can easily
                    # make the buttons navigate to other pages, without
                    # having to write an event handler. Notice how there is
                    # no Python function called when the button is clicked.
                    rio.Link(
                        rio.Button(
                            "News",
                            icon="material/news",
                            style=(
                                "major"
                                if active_page_url_segment == "news-page"
                                else "plain-text"
                            ),
                        ),
                        "/news-page",
                    ),
                    # Same game, different button
                    rio.Link(
                        rio.Button(
                            "About",
                            icon="material/info",
                            style=(
                                "major"
                                if active_page_url_segment == "about-page"
                                else "plain-text"
                            ),
                        ),
                        "/about-page",
                    ),
                    spacing=1,
                    margin=1,
                ),
                # Set the fill of the rectangle to the neutral color of the theme and
                # Add a corner radius
                fill=self.session.theme.neutral_color,
                corner_radius=self.session.theme.corner_radius_medium,
                # Add shadow properties
                shadow_radius=0.8,
                shadow_color=self.session.theme.shadow_color,
                shadow_offset_y=0.2,
                # Overlay assigns the entire screen to its child component.
                # Since the navbar isn't supposed to take up all space, assign
                # an alignment.
                align_y=0,
                margin_x=5,
                margin_y=2,
            )
        )


# </component>



from __future__ import annotations

import copy
import functools
import json
import re
import typing as t
import urllib.parse
from dataclasses import dataclass
from pathlib import Path

import typing_extensions as te
import uniserde

from .. import utils

SECTION_PATTERN = re.compile(r" *#\s*<(\/?[\w-]+)>")


# Contains default values for the `meta.json` file
DEFAULT_META_DICT = {
    "dependencies": {},
    "rootComponent": None,
    "onAppStart": None,
    "onSessionStart": None,
    "defaultAttachments": None,
}


# All Available templates, including the empty template
#
# THE ORDER MATTERS. `revel` will display the options in the same order as they
# appear in the literal
AvailableTemplatesLiteral: te.TypeAlias = t.Literal[
    # Keep the empty template first
    "Empty",
    # Sort the remainder alphabetically
    "AI Chatbot",
    "Authentication",
    "Crypto Dashboard",
    "Multipage Website",
    "Simple CRUD",
    "Tic-Tac-Toe",
    "Todo App",
]


@dataclass
class _TemplateConfig(uniserde.Serde):
    """
    Model for parsing the JSON file which comes along with each project
    template.
    """

    # Allows displaying the templates in a structured way
    level: t.Literal["beginner", "intermediate", "advanced"]

    # Very short, one or two line description of the template
    summary: str

    # Any pypi packages this project depends on, in the same format as used by
    # `pip`.
    #
    # Example: `{"numpy": ">=1.21.0", "pandas": "^1.3.0"}`
    dependencies: dict[str, str]

    # Additional parameters to pass to the app instance
    root_component: str | None
    on_app_start: str | None
    on_session_start: str | None
    default_attachments: list[str] | None


@dataclass
class Snippet:
    # The group the snippet is in
    group: str

    # The name the snippet can be accessed with
    name: str

    # The path to the snippet file
    file_path: Path

    # For text snippets, this is the raw file contents. `None` for binary
    # snippets.
    raw_code: str | None

    @property
    def is_text_snippet(self) -> bool:
        return self.raw_code is not None

    @property
    def is_binary_snippet(self) -> bool:
        return not self.is_text_snippet

    @staticmethod
    def from_path(group: str, name: str, file_path: Path) -> Snippet:
        # Read the contents if it's a text snippet
        if file_path.suffix in (".txt", ".md", ".py", ".json"):
            raw_code = file_path.read_text(encoding="utf-8")

            # The example projects contain a bunch of imports that can't be
            # resolved, so those are marked with special `# type: ignore`
            # comments that need to be removed
            if file_path.suffix == ".py":
                raw_code = raw_code.replace(
                    "  # type: ignore (hidden from user)", ""
                )
        else:
            raw_code = None

        return Snippet(
            group=group,
            name=name,
            file_path=file_path,
            raw_code=raw_code,
        )

    def get_section(self, section_name: str) -> str:
        assert self.is_text_snippet, self
        assert self.raw_code is not None, self  # For the type checker

        # Find the target section
        lines = self.raw_code.splitlines()
        section_found = False
        result = []

        for line in lines:
            match = SECTION_PATTERN.match(line)

            # No match
            if match is None:
                result.append(line)
                continue

            # Start of the target section?
            if match.group(1) == section_name:
                result = []
                section_found = True
                continue

            # End of the target section?
            if match.group(1) == f"/{section_name}":
                return "\n".join(result)

            # Some other section, drop it
            pass

        # The section was never found
        if not section_found:
            raise KeyError(f"There is no section named `{section_name}`")

        # The section was found, but never closed
        raise ValueError(f"The section `{section_name}` was never closed")

    def stripped_code(self) -> str:
        """
        Returns the given code with all section tags removed.
        """
        assert self.is_text_snippet, self
        assert self.raw_code is not None, self  # For the type checker

        lines = self.raw_code.splitlines()

        # Remove all section tags
        result = []
        for line in lines:
            if SECTION_PATTERN.match(line) is None:
                result.append(line)

        return "\n".join(result)


@functools.lru_cache(maxsize=None)
def _get_all_snippet_paths() -> dict[str, dict[str, Path]]:
    """
    Returns a dictionary containing all available snippets. The snippets are
    organized by group and name.

    The snippets are read from the `snippets` directory when the function is
    first called. The result is cached after that. Because of this, the result
    mustn't be modified.
    """
    # Find all snippet files
    result = {}

    def scan_dir_recursively(group_name: str, path: Path) -> None:
        assert path.is_dir(), path
        assert result is not None

        # Exclude some directories
        if path.name == "__pycache__":
            return

        for fpath in path.iterdir():
            # Directory
            if fpath.is_dir():
                scan_dir_recursively(group_name, fpath)

            # Snippet file
            else:
                name = fpath.name
                group = result.setdefault(group_name, {})
                group[name] = fpath

    # Scan all snippet directories. The first directory is used as a key, the
    # rest just for organization.
    for group_dir in utils.SNIPPETS_DIR.iterdir():
        assert group_dir.is_dir(), group_dir
        scan_dir_recursively(group_dir.name, group_dir)

    return result


def get_snippet_groups() -> set[str]:
    """
    Returns a set of all snippet available groups.
    """
    all_groups = _get_all_snippet_paths()
    return set(all_groups.keys())


@functools.lru_cache(maxsize=None)
def all_snippets_in_group(group: str) -> t.Iterable[Snippet]:
    """
    Returns all snippets in the given group.

    ## Raises

    `KeyError`: If there is no group with the given name.
    """
    all_groups = _get_all_snippet_paths()
    group_dict = all_groups.get(group, {})

    return tuple(
        Snippet.from_path(group, name, file_path)
        for name, file_path in group_dict.items()
    )


@functools.lru_cache(maxsize=None)
def get_snippet(group: str, name: str) -> Snippet:
    """
    Parses and returns the snippet with the given group and name.

    ## Raises

    `KeyError`: If there is no snippet with the given group and name.
    """
    all_groups = _get_all_snippet_paths()

    group_dict = all_groups[group]
    file_path = group_dict[name]

    return Snippet.from_path(group, name, file_path)


@dataclass
class ProjectTemplate:
    """
    Project templates are stored as snippets. This class represents all
    resources necessary for a single project template.
    """

    # Human-readable name of the project template
    name: AvailableTemplatesLiteral

    # How difficult the project is
    level: t.Literal["beginner", "intermediate", "advanced"]

    # A short description of the project template
    summary: str

    # Description of the project template (Markdown)
    description_markdown_source: str

    # The project thumbnail. This is a SVG file.
    thumbnail: Snippet

    # Any pypi packages this project depends on, in the same format as used by
    # `pip`.
    #
    # Example: `{"numpy": ">=1.21.0", "pandas": "^1.3.0"}`
    dependencies: dict[str, str]

    # All snippets which should be included
    asset_snippets: list[Snippet]
    component_snippets: list[Snippet]
    page_snippets: list[Snippet]
    other_python_files: list[Snippet]

    root_init_snippet: Snippet

    # Additional configuration for the app instance
    root_component: str | None
    on_app_start: str | None
    on_session_start: str | None
    default_attachments: list[str] | None

    @property
    def slug(self) -> str:
        return self.name.lower().replace(" ", "-")

    @staticmethod
    def _from_snippet_group(
        snippet_name: str,
        snippets: t.Iterable[Snippet],
    ) -> ProjectTemplate:
        assert (
            snippet_name in t.get_args(AvailableTemplatesLiteral)
            or snippet_name == "Empty"
        ), snippet_name
        name = t.cast(AvailableTemplatesLiteral, snippet_name)

        # Find all snippets needed for the project template
        readme_snippet: Snippet | None = None
        thumbnail_snippet: Snippet | None = None
        metadata: _TemplateConfig | None = None
        root_init_snippet: Snippet | None = None

        asset_snippets: list[Snippet] = []
        python_files: list[Snippet] = []

        for snippet in snippets:
            # README snippet can be recognized by its name
            if snippet.name == "README.md":
                assert readme_snippet is None
                assert snippet.is_text_snippet, snippet.file_path
                readme_snippet = snippet
                continue

            # As is the thumbnail
            if snippet.name == "thumbnail.svg":
                assert thumbnail_snippet is None
                thumbnail_snippet = snippet
                continue

            # And the metadata
            if snippet.name == "meta.json":
                meta_dict: dict[str, t.Any] = copy.deepcopy(DEFAULT_META_DICT)
                meta_dict.update(json.loads(snippet.stripped_code()))
                metadata = _TemplateConfig.from_json(meta_dict)
                continue

            # Others are categorized by the directory they're in
            dir_name = snippet.file_path.parent.name

            if dir_name == "assets":
                asset_snippets.append(snippet)

            elif snippet.file_path.name == "root_init.py":
                assert root_init_snippet is None
                assert snippet.is_text_snippet, snippet.file_path
                root_init_snippet = snippet

            elif snippet.file_path.suffix == ".py":
                python_files.append(snippet)

            else:
                assert False, f"Unrecognized snippet file `{snippet.file_path}`"

        # Make sure everything was found
        assert (
            readme_snippet is not None
        ), f"`README.md` snippet not found for `{name}`"
        assert (
            thumbnail_snippet is not None
        ), f"`thumbnail.svg` snippet not found for {name}"
        assert (
            metadata is not None
        ), f"`meta.json` snippet not found for `{name}`"
        assert (
            root_init_snippet is not None
        ), f"`root_init.py` snippet not found for `{name}`"

        # Further split the Python files into components, pages, and other
        # files. This cannot be done above, because it requires knowledge of
        # where the snippet groups's root directory is placed. This directory in
        # turn is only known once the snippets have been found.
        template_root_dir = root_init_snippet.file_path.parent

        ii = 0
        component_snippets: list[Snippet] = []
        page_snippets: list[Snippet] = []

        while ii < len(python_files):
            snippet = python_files[ii]
            rel_path = snippet.file_path.relative_to(template_root_dir)

            # Component?
            if rel_path.parts[0] == "components":
                del python_files[ii]

                if snippet.name == "__init__.py":
                    continue

                assert snippet.is_text_snippet, snippet.file_path
                component_snippets.append(snippet)

            # Page?
            elif rel_path.parts[0] == "pages":
                del python_files[ii]

                assert snippet.is_text_snippet, snippet.file_path
                page_snippets.append(snippet)

            # Plain old Python file
            else:
                ii += 1

        assert len(page_snippets) > 0, f"No page snippets found for `{name}`"

        # Create the project template
        return ProjectTemplate(
            name=name,
            level=metadata.level,
            summary=metadata.summary,
            description_markdown_source=readme_snippet.stripped_code(),
            thumbnail=thumbnail_snippet,
            dependencies=metadata.dependencies,
            asset_snippets=asset_snippets,
            component_snippets=component_snippets,
            page_snippets=page_snippets,
            other_python_files=python_files,
            root_init_snippet=root_init_snippet,
            root_component=metadata.root_component,
            on_app_start=metadata.on_app_start,
            on_session_start=metadata.on_session_start,
            default_attachments=metadata.default_attachments,
        )


@functools.lru_cache(maxsize=None)
def get_project_templates(include_empty: bool) -> t.Iterable[ProjectTemplate]:
    """
    Iterates over all available project templates.

    `include_empty` controls whether to include the empty template.
    """
    result = []

    # Templates are just snippet groups starting with `project-template-`
    for group_name, snippets in _get_all_snippet_paths().items():
        # Is this a project template?
        template_name = group_name.removeprefix("project-template-")

        if group_name == template_name:
            continue

        # Skip the empty template if it's not requested
        if template_name == "Empty" and not include_empty:
            continue

        # Parse the template
        result.append(
            ProjectTemplate._from_snippet_group(
                template_name,
                [
                    Snippet.from_path(
                        group_name,
                        name,
                        file_path,
                    )
                    for name, file_path in snippets.items()
                ],
            )
        )

    # This function can't simply yield, because of the `lru_cache`. Instead,
    # return something immutable.
    return tuple(result)


@dataclass
class HowtoGuide:
    """
    Represents a how-to guide.
    """

    # A URL-safe unique identifier for the how-to guide
    slug: str

    # A human-readable name of the how-to guide
    title: str

    # The markdown source of the how-to guide. This does not contain a title
    markdown_source: str

    @staticmethod
    def from_snippet(snip: Snippet) -> HowtoGuide:
        """
        Creates a `Howto` instance from a snippet. The snippet must contain
        the howto's markdown source and start with a # heading.
        """
        # Split the markdown source into title and content
        full_source = snip.stripped_code()
        lines = full_source.split("\n", 1)
        assert (
            len(lines) == 2
        ), f"Snippet `{snip.name}` is either missing a title or the content"
        title, content = lines

        # Clean up the title
        title = title.strip()
        assert title.startswith(
            "#"
        ), f"The source for snippet `{snip.name}` does not start with a heading?"
        title = title[1:].strip()

        # Clean up the content
        content = content.strip()

        # Derive the slug from the snippet's name
        assert snip.name.endswith(".md"), snip.name
        slug = snip.name[:-3]
        assert slug == urllib.parse.quote(slug), slug

        # Build the result
        return HowtoGuide(
            slug=slug,
            title=title,
            markdown_source=content,
        )


@functools.lru_cache(maxsize=None)
def get_howto_guides() -> t.Iterable[HowtoGuide]:
    """
    Iterates over all available how-to guides.
    """
    result = []

    for snip in all_snippets_in_group("howtos"):
        result.append(HowtoGuide.from_snippet(snip))

    # This function can't simply yield, because of the `lru_cache`. Instead,
    # return something immutable.
    return tuple(result)



from __future__ import annotations

# <additional-imports>
from datetime import datetime, timezone

import rio

from .. import data_models, persistence

# </additional-imports>


# <component>
class Navbar(rio.Component):
    """
    A navbar with a fixed position and responsive width.
    """

    # Make sure the navbar will be rebuilt when the app navigates to a different
    # page. While Rio automatically detects state changes and rebuilds
    # components as needed, navigating to other pages is not considered a state
    # change, since it's not stored in the component.
    #
    # Instead, we can use Rio's `on_page_change` event to trigger a rebuild of
    # the navbar when the page changes.
    @rio.event.on_page_change
    async def on_page_change(self) -> None:
        # Rio comes with a function specifically for this. Whenever Rio is
        # unable to detect a change automatically, use this function to force a
        # refresh.
        await self.force_refresh()

    async def on_logout(self) -> None:
        user_session = self.session[data_models.UserSession]

        # Expire the session
        pers = self.session[persistence.Persistence]

        await pers.update_session_duration(
            user_session,
            new_valid_until=datetime.now(tz=timezone.utc),
        )

        # Detach everything from the session. This informs all components that
        # nobody is logged in.
        self.session.detach(data_models.AppUser)
        self.session.detach(data_models.UserSession)

        # Navigate to the login page to prevent the user being on a page that is
        # prohibited without being logged in.
        self.session.navigate_to("/")

    def build(self) -> rio.Component:
        try:
            # Which page is currently active? This will be used to highlight the
            # correct navigation button.
            #
            # `active_page_instances` contains `rio.ComponentPage` instances
            # that that are created during app creation. Since multiple pages
            # can be active at a time (e.g. /foo/bar/baz), this is a list rather
            # than just a single page.
            active_page = self.session.active_page_instances[1]
            active_page_url_segment = active_page.url_segment
        except IndexError:
            # Handle the case where there are no active sub-pages. e.g. when the
            # user is not logged in.
            active_page_url_segment = None

        # Check if the user is logged in and display the appropriate buttons
        # based on the user's status
        try:
            self.session[data_models.AppUser]

        # If no user is attached, nobody is logged in
        except KeyError:
            user_settings = False

        # If a user is attached, they are logged in
        else:
            user_settings = True

        # Create the content of the navbar. First we create a row with a certain
        # spacing and margin.  We can use the `.add()` method to add components
        # by condition to the row.
        navbar_content = rio.Row(spacing=1, margin=1)

        # Links can be used to navigate to other pages and
        # external URLs. You can pass either a simple string, or
        # another component as their content.
        navbar_content.add(
            rio.Link(
                rio.IconButton(
                    "rio/logo",
                    style="plain-text",
                    min_size=2.5,
                ),
                "/app/home",
            )
        )

        # This spacer will take up any superfluous space,
        # effectively pushing the subsequent buttons to the
        # right.
        navbar_content.add(rio.Spacer())

        # Based on the user's status, display the appropriate buttons
        if user_settings:
            # By sticking buttons into a `rio.Link`, we can easily
            # make the buttons navigate to other pages, without
            # having to write an event handler. Notice how there is
            # no Python function called when the button is clicked.
            navbar_content.add(
                rio.Link(
                    rio.Button(
                        "Home",
                        icon="material/news",
                        style=(
                            "major"
                            if active_page_url_segment == "home"
                            else "plain-text"
                        ),
                    ),
                    "/app/home",
                )
            )

            # Same game, different button
            navbar_content.add(
                rio.Link(
                    rio.Button(
                        "News",
                        icon="material/news",
                        style=(
                            "major"
                            if active_page_url_segment == "news-page"
                            else "plain-text"
                        ),
                    ),
                    "/app/news-page",
                )
            )

            # Same game, different button
            navbar_content.add(
                rio.Link(
                    rio.Button(
                        "About",
                        icon="material/info",
                        style=(
                            "major"
                            if active_page_url_segment == "about-page"
                            else "plain-text"
                        ),
                    ),
                    "/app/about-page",
                )
            )

            # Logout
            navbar_content.add(
                rio.Button(
                    "Logout",
                    icon="material/logout",
                    style="plain-text",
                    on_press=self.on_logout,
                )
            )

        # Display the login button if the user is not logged in
        else:
            navbar_content.add(
                rio.Link(
                    rio.Button(
                        "Login",
                        icon="material/login",
                        style="plain-text",
                    ),
                    "/",
                )
            )

        # The navbar should appear above all other components. This is easily
        # done by using a `rio.Overlay` component.
        return rio.Overlay(
            # Use a rectangle for visual separation
            rio.Rectangle(
                # Use the content we've built up by conditions
                content=navbar_content,
                # Set the fill of the rectangle to the neutral color of the
                # theme
                fill=self.session.theme.neutral_color,
                # Round the corners
                corner_radius=self.session.theme.corner_radius_medium,
                # Add a shadow to make the navbar stand out above other content
                shadow_radius=0.8,
                shadow_color=self.session.theme.shadow_color,
                shadow_offset_y=0.2,
                # Overlay assigns the entire screen to its child component.
                # Since the navbar isn't supposed to take up all space, align
                # it.
                align_y=0,
                margin_x=5,
                margin_y=2,
            ),
        )


# </component>



from __future__ import annotations

import enum
import typing as t
from datetime import timedelta

__all__ = [
    "on_mount",
    "on_page_change",
    "on_populate",
    "on_unmount",
    "on_window_size_change",
    "periodic",
]


R = t.TypeVar("R")
SyncOrAsync = R | t.Awaitable[R]
SyncOrAsyncNone = t.TypeVar("SyncOrAsyncNone", bound=SyncOrAsync[None])

Func = t.TypeVar("Func", bound=t.Callable)
Decorator = t.Callable[[Func], Func]

MethodWithNoParametersVar = t.TypeVar(
    "MethodWithNoParametersVar", bound=t.Callable[[t.Any], t.Any]
)


class EventTag(enum.Enum):
    """
    Event tags are internal markers used to keep track of which function needs
    to be called when.

    ## Metadata

    `public`: False
    """

    ON_MOUNT = enum.auto()
    ON_PAGE_CHANGE = enum.auto()
    ON_POPULATE = enum.auto()
    ON_UNMOUNT = enum.auto()
    ON_WINDOW_SIZE_CHANGE = enum.auto()
    PERIODIC = enum.auto()


def _tag_as_event_handler(
    function: t.Callable,
    tag: EventTag,
    args: t.Any,
) -> None:
    """
    Registers the function as an event handler for the given tag. This simply
    adds a marker to the function's `__dict__` so that it can later be
    recognized as an event handler.
    """
    all_events: dict[EventTag, list[t.Any]] = vars(function).setdefault(
        "_rio_events_", {}
    )
    events_like_this = all_events.setdefault(tag, [])
    events_like_this.append(args)


def on_mount(handler: MethodWithNoParametersVar) -> MethodWithNoParametersVar:
    """
    Triggered when the component is added to the component tree.

    This decorator makes the decorated method an event handler for `on_mount`
    events. The method will be called whenever the component is added to the
    component tree.

    This may be triggered multiple times if the component is removed and then
    re-added.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Example

    Here's an example of a component being conditionally included in the
    component tree. The `Switch` controls whether the `OnMountPrinter` exists or
    not, so turning on the switch will mount the `OnMountPrinter` and print
    "Mounted" to the console.

    ```python
    class OnMountPrinter(rio.Component):
        @rio.event.on_mount
        def on_mount(self):
            print("Mounted")

        def build(self):
            return rio.Text("hello")


    class Toggler(rio.Component):
        child: rio.Component
        show_child: bool = False

        def build(self) -> rio.Component:
            return rio.Column(
                # Depending on the Switch state, show either the
                # child or a placeholder
                self.child if self.show_child else rio.Text(""),
                rio.Switch(is_on=self.bind().show_child),
            )


    app = rio.App(build=lambda: Toggler(OnMountPrinter()))
    app.run_in_browser()
    ```


    ## Metadata

    `decorator`: True
    """
    _tag_as_event_handler(handler, EventTag.ON_MOUNT, None)
    return handler


def on_page_change(
    handler: MethodWithNoParametersVar,
) -> MethodWithNoParametersVar:
    """
    Triggered whenever the session changes pages.

    Makes the decorated method an event handler for `on_page_change` events.
    The method will be called whenever the session navigates to a new page.

    If you want your code to run both when the component was first created _and_
    when the page changes, you can combine this decorator with `__post_init__`
    or `on_populate`.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Example

    ```python
    class UrlDisplay(rio.Component):
        current_url: rio.URL = rio.URL()

        def __post_init__(self):
            self.current_url = self.session.active_page_url

        @rio.event.on_page_change
        def on_page_change(self):
            self.current_url = self.session.active_page_url

        def build(self):
            return rio.Text(f"You're currently on {self.current_url}")
    ```


    ## Metadata

    `decorator`: True
    """
    _tag_as_event_handler(handler, EventTag.ON_PAGE_CHANGE, None)
    return handler


def on_populate(
    handler: MethodWithNoParametersVar,
) -> MethodWithNoParametersVar:
    """
    Triggered when the component has been created or has been reconciled.

    This decorator makes the decorated method an event handler for `on_populate`
    events. The method will be called whenever the component has been created or
    has been reconciled. This allows you to asynchronously fetch any data right
    after component initialization.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Example

    `on_populate` is often useful as a sort of "async init", where you can put
    async code that needs to be run when the component is instantiated. In this
    example we'll use it to perform an asynchronous HTTP request.

    ```python
    import httpx
    from dataclasses import field


    class PypiVersionFetcher(rio.Component):
        module: str
        # The `version` attribute is initialized to an empty
        # string, which will will act as a placeholder until
        # the HTTP request finishes
        version: str = field(init=False, default="")

        @rio.event.on_populate
        async def on_populate(self):
            # Whenever this component is instantiated or
            # reconciled, fetch the version of the given module
            async with httpx.AsyncClient() as client:
                url = f"https://pypi.org/pypi/{self.module}/json"
                response = await client.get(url)
                self.version = response.json()["info"]["version"]

        def build(self):
            return rio.Text(f"Latest {self.module} version: {self.version}")
    ```


    ## Metadata

    `decorator`: True
    """

    _tag_as_event_handler(handler, EventTag.ON_POPULATE, None)
    return handler


def on_unmount(handler: MethodWithNoParametersVar) -> MethodWithNoParametersVar:
    """
    Triggered when the component is removed from the component tree.

    This decorator makes the decorated method an event handler for `on_unmount`
    events. The method will be called whenever the component is removed from the
    component tree.

    This may be triggered multiple times if the component is removed and then
    re-added.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Example

    Here's an example of a component being conditionally included in the
    component tree. The `Switch` controls whether the `OnUnmountPrinter` exists
    or not, so turning on the switch will mount the `OnUnmountPrinter` and print
    "Unmounted" to the console.

    ```python
    class OnUnmountPrinter(rio.Component):
        @rio.event.on_unmount
        def on_unmount(self):
            print("Unmounted")

        def build(self):
            return rio.Text("hello")


    class Toggler(rio.Component):
        child: rio.Component
        show_child: bool = False

        def build(self) -> rio.Component:
            return rio.Column(
                # Depending on the Switch state, show either the
                # child or a placeholder
                self.child if self.show_child else rio.Text(""),
                rio.Switch(is_on=self.bind().show_child),
            )


    app = rio.App(build=lambda: Toggler(OnUnmountPrinter()))
    app.run_in_browser()
    ```


    ## Metadata

    `decorator`: True
    """
    _tag_as_event_handler(handler, EventTag.ON_UNMOUNT, None)
    return handler


def on_window_size_change(
    handler: MethodWithNoParametersVar,
) -> MethodWithNoParametersVar:
    """
    Triggered when the client's window is resized.

    This decorator makes the decorated method an event handler for
    `on_window_size_change` events. The method will be called whenever the
    window changes size. You can access the window's size using
    `self.session.window_width` and `self.session.window_height` as usual.

    Some of the ways that a window can resize are non obvious. For example,
    rotating a mobile device will trigger this event, since width and height
    trade places. This event may also be triggered when the browser's dev tools
    are opened or closed, or when the browser's zoom level is changed, since all
    of those impact the available screen space.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Example

    We'll make a component that displays the size of the window. The
    `@window_size_change` event is used to rebuild the component whenever the
    window size changes. (This doesn't happen automatically because Rio only
    rebuilds components when their attributes change, and this component doesn't
    have any attributes that change.)

    ```python
    class WindowSizeDisplay(rio.Component):
        @rio.event.on_window_size_change
        async def on_window_size_change(self):
            await self.force_refresh()

        def build(self):
            width = self.session.window_width
            height = self.session.window_height
            return rio.Text(f"The window size is {width:.1f}x{height:.1f}")
    ```


    ## Metadata

    `decorator`: True
    """
    _tag_as_event_handler(handler, EventTag.ON_WINDOW_SIZE_CHANGE, None)
    return handler


def periodic(
    interval: float | timedelta,
) -> Decorator[MethodWithNoParametersVar]:
    """
    Triggered at a fixed time interval.

    This decorator causes the decorated method to be called repeatedly at a
    fixed time interval. The event will be triggered for as long as the
    component exists, even if it is not mounted. The interval can be specified
    as either a number of seconds or as a timedelta.

    The interval only starts counting after the previous handler has finished
    executing, so the handler will never run twice simultaneously, even if it
    takes longer than the interval to execute.

    This decorator can be used on both synchronous as well as asynchronous
    methods.


    ## Parameters

    `interval`: The number of seconds, or timedelta, between each trigger.


    ## Example

    Here we use `@rio.event.periodic` to increment a counter every second:

    ```python
    class Counter(rio.Component):
        count: int = 0

        @rio.event.periodic(1)
        def increment_count(self):
            self.count += 1

        def build(self):
            return rio.Text(f"{self.count} seconds have passed")
    ```


    ## Metadata

    `decorator`: True
    `experimental`: True
    """
    # Convert timedelta to float
    if isinstance(interval, timedelta):
        interval = interval.total_seconds()

    def decorator(
        handler: MethodWithNoParametersVar,
    ) -> MethodWithNoParametersVar:
        _tag_as_event_handler(handler, EventTag.PERIODIC, interval)
        return handler

    return decorator



# <additional-imports>
import pandas as pd
import plotly.express as px

import rio

from .. import data_models

# </additional-imports>


# <component>
class BalanceCard(rio.Component):
    """
    Displays information about the user's balance.

    You can build Components also in a functional-oriented way. Here we have
    two functions that create all sections of the balance card:
    - balance_section
    - bar_chart_section

    The build method combines these sections into a single `rio.Card` component,
    creating a complete balance component for the dashboard.

    ## Attributes

    `data`: Historical data of the fetched crypto coins.
    """

    data: pd.DataFrame

    def total_balance(self, idx: int) -> float:
        """
        Calculates the total amount of our portfolio in USD.

        This function iterates over the coins in MY_COINS, and for each coin, it
        multiplies the owned coin amount of our portfolio with the current USD
        exchange rate. The sum of our portfolio in USD will be returned.


        ## Parameters

        `idx`: The index at which to calculate the total balance.
        """

        total = 0

        for coin in data_models.MY_COINS:
            total += data_models.MY_COINS[coin][0] * self.data[coin].iloc[idx]

        return total

    def percentual_differance_balance(self) -> float:
        """
        Calculates percentual change of our portfolio.
        """

        total_last = 0
        total_second_last = 0
        epsilon = 0.0000001

        for coin in data_models.MY_COINS:
            total_last += data_models.MY_COINS[coin][0] * self.total_balance(
                idx=-1
            )
            total_second_last += data_models.MY_COINS[coin][
                0
            ] * self.total_balance(idx=-2)

        # epsilon ensures that the denominator is never zero
        return (
            (total_last - total_second_last) / (total_second_last + epsilon)
        ) * 100

    def balance_section(self) -> rio.Component:
        """
        Creates a balance section for the dashboard.

        A section that displays the total balance and the percental difference
        in balance. The total balance is displayed in bold, and the percental
        difference is displayed in green if it's positive and in red if it's
        negative. The section is returned as a Column component.

        ```
        ╔════════════════════════ Column ════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ My Balance                                         ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Total Balance                                      ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏━━━━━━━━ Text ━━━━━━━━┓  ┏━━━━━━━━ Text ━━━━━━━━┓ ┃ ║
        ║ ┃ ┃ e.g. 400,000 USD     ┃  ┃ e.g. (-0.09 %)       ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━━━━━━┛  ┗━━━━━━━━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚════════════════════════════════════════════════════════╝
        ```
        """

        return rio.Column(
            rio.Text(
                "My Balance",
                justify="center",
                style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                align_x=0,
                margin_bottom=1,
            ),
            rio.Text(
                "Total Balance",
                justify="center",
                style="dim",
                align_x=0,
            ),
            rio.Row(
                rio.Text(
                    f"{self.total_balance(idx=-1):,.2f} USD",
                    justify="center",
                    style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                    align_x=0,
                ),
                rio.Text(
                    f"({self.percentual_differance_balance():.2f} %)",
                    justify="center",
                    style=rio.TextStyle(
                        fill=(
                            rio.Color.GREEN
                            if self.percentual_differance_balance() > 0
                            else rio.Color.RED
                        )
                    ),
                ),
                spacing=1,
            ),
            spacing=1,
            align_y=1,
        )

    def chart_section(self, name: str, color: str) -> rio.Component:
        """
        Creates a bar chart section for the dashboard.

        ## Parameters

        `name`: The name of the section.

        `color`: The color of the bars in the bar chart.

        ```
        ╔════════════════════════ Column ════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━ Plot ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Bar Chart                                          ┃ ║
        ║ ┃                                                    ┃ ║
        ║ ┃                                                    ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ Income | Expenses                                  ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━ Text ━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ e.g. 980,000.00 USD                                ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚════════════════════════════════════════════════════════╝
        ```
        """

        fig = px.bar(
            data_models.BAR_CHART,
            height=200,
            width=200,
            color_discrete_sequence=[color],
        )
        # hide and lock down axes
        fig.update_xaxes(visible=False, fixedrange=True)
        fig.update_yaxes(visible=False, fixedrange=True)
        # remove facet/subplot labels
        fig.update_layout(annotations=[], overwrite=True)

        # strip down the rest of the plot
        fig.update_layout(
            showlegend=False,
            margin=dict(t=10, l=10, b=10, r=10),
        )

        return rio.Column(
            rio.Plot(
                figure=fig,
                min_height=5,
                min_width=20,
                background=self.session.theme.neutral_color,
            ),
            rio.Text(
                name,
                justify="center",
                style="dim",
                align_x=0,
            ),
            rio.Text(
                f"{self.total_balance(idx=-1):,.2f} USD",
                style=rio.TextStyle(font_size=1.2, font_weight="bold"),
                justify="left",
            ),
            spacing=1,
            align_y=1,
        )

    def build(self) -> rio.Component:
        """
        Returns a complete balance card component for the dashboard.

        The balance card component consists of three sections and seperators
        between them:
        - balance_section
        - chart_section

        The sections are combined into a single `rio.Card` component, which is
        returned as the final balance card component for the dashboard.

        ```
        ╔══════════════════════════════ Card ═════════════════════════════╗
        ║ ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━ Row ━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓ ║
        ║ ┃ ┏ balance_section ┓ ║ ┏ chart_section ┓ ║ ┏ chart_section ┓ ┃ ║
        ║ ┃ ┃                 ┃ ║ ┃               ┃ ║ ┃               ┃ ┃ ║
        ║ ┃ ┗━━━━━━━━━━━━━━━━━┛ ║ ┗━━━━━━━━━━━━━━━┛ ║ ┗━━━━━━━━━━━━━━━┛ ┃ ║
        ║ ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛ ║
        ╚═════════════════════════════════════════════════════════════════╝
        """

        return rio.Card(
            rio.Row(
                # 1. Section
                self.balance_section(),
                rio.Separator(),
                # 2. Section
                self.chart_section(name="Income", color="green"),
                rio.Separator(),
                self.chart_section(name="Expenses", color="red"),
                margin=1,
                spacing=4,
                align_x=0.5,
            ),
            min_height=13,
        )


# </component>



"""
Contains documentation related tasks specific to the Rio project.
"""

import dataclasses
import functools
import inspect
import re
import types
import typing as t

import imy.docstrings
import introspection
import unicall

import rio

__all__ = [
    "mark_as_private",
    "mark_constructor_as_private",
    "get_docs_for",
    "get_documentation_fragment",
    "build_documentation_url",
    "find_documented_objects",
    "insert_links_into_markdown",
]


Class = t.TypeVar("Class", bound=type)
ClassOrFunction = t.TypeVar("ClassOrFunction", bound=type | types.FunctionType)


_NAME_TO_URL: dict[str, str] | None = None
_CODE_BLOCK_REGEX = re.compile(r"(.*?)(```.*?```|$)", flags=re.S)
_AUTO_LINK_REGEX = re.compile(r"`(?:rio\.)?([a-zA-Z_.]+)`(?!\])")


def insert_links_into_markdown(
    markdown: str, *, own_name: str | None = None
) -> str:
    """
    Turns markdown like

        `Row`

    into a link like

        [`Row`](/docs/api/row)

    The `own_name` parameter can be used to prevent a page from linking to
    itself. For example, in the documentation for `Row`, we don't want every
    occurrence of `Row` to turn into a link.
    """
    global _NAME_TO_URL

    if _NAME_TO_URL is None:
        _NAME_TO_URL = {}

        for docs in _get_unprocessed_docs().values():
            url = str(build_documentation_url(docs.name))
            _NAME_TO_URL[docs.name] = url

            # TODO: Also link methods and attributes (and parameters?) once they
            # have #url-fragments

    name_to_url = _NAME_TO_URL  # Reassign to appease the static type checker

    def repl(match: re.Match) -> str:
        name = match.group(1)

        if name != own_name:
            try:
                url = name_to_url[name]
            except KeyError:
                pass
            else:
                return f"[{match.group()}]({url})"

        return match.group()

    # We want to look for single-line text like `Row` and turn it into a link.
    # The problem is that such text might be appear inside of a code block (in a
    # comment). So we'll search for code blocks, and only apply the
    # substituation in the text before the code block.
    chunks = list[str]()

    for match_ in _CODE_BLOCK_REGEX.finditer(markdown):
        text, code_block = match_.groups()

        chunks.append(_AUTO_LINK_REGEX.sub(repl, text))
        chunks.append(code_block)

    return "".join(chunks)


def mark_as_private(obj: ClassOrFunction) -> ClassOrFunction:
    if obj.__doc__ is None:
        obj.__doc__ = "## Metadata"
    elif "## Metadata" not in obj.__doc__:
        obj.__doc__ += "\n## Metadata"

    obj.__doc__ += "\n`public`: False"

    return obj


def mark_constructor_as_private(cls: Class) -> Class:
    try:
        constructor = vars(cls)["__init__"]
    except KeyError:

        def constructor(self, *args, **kwargs):
            super(cls, self).__init__(*args, **kwargs)  # type: ignore (wtf?)

        introspection.add_method_to_class(constructor, cls, "__init__")

    mark_as_private(constructor)  # type: ignore (wtf?)

    return cls


@t.overload
def get_docs_for(obj: type) -> imy.docstrings.ClassDocs: ...


@t.overload
def get_docs_for(obj: types.FunctionType) -> imy.docstrings.FunctionDocs: ...


def get_docs_for(
    obj: t.Callable | t.Type,
) -> imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs:
    """
    Parse the docs for a component and return them. The results are cached, so
    this function is fast.
    """
    return find_documented_objects()[obj]


def _make_docs_for_rio_url():
    docs = imy.docstrings.ClassDocs.from_class(rio.URL)
    docs.attributes.clear()
    docs.functions.clear()
    docs.summary = "Alias for `yarl.URL`."
    docs.details = """
Since URLs are a commonly used data type, `rio` re-exports `yarl.URL` as
`rio.URL` for convenience. See the
[`yarl` documentation](https://yarl.aio-libs.org/en/stable/api/#yarl.URL) for
details about this class.
""".strip()

    return docs


def get_documentation_fragment(object_name: str) -> str:
    """
    Returns the fragment part of the URL corresponding to the documentation for
    the given Rio object.
    """
    return object_name.lower()


def build_documentation_url(
    object_name: str,
    *,
    relative: bool = False,
) -> rio.URL:
    """
    Returns the URL to the documentation for the given Rio object. This doesn't
    perform any checks on whether the object actually exists and has
    documentation. It relies solely on the passed values.
    """

    # Build the relative URL
    result_string = f"/docs/api/{get_documentation_fragment(object_name)}"

    # Make it absolute, if requested
    if not relative:
        result_string = "https://rio.dev" + result_string

    # Done
    return rio.URL(result_string)


def _find_possibly_public_objects() -> t.Iterable[t.Type | t.Callable]:
    """
    Finds all objects in rio that might be public. This uses heuristics to
    filter out many internal objects, but it's not perfect.
    """
    # Hardcoded items
    yield rio.App
    yield rio.AssetError
    yield rio.Color
    yield rio.ComponentPage
    yield rio.CursorStyle
    yield rio.escape_markdown
    yield rio.escape_markdown_code
    yield rio.FileInfo
    yield rio.Font
    yield rio.NavigationFailed
    yield rio.page
    yield rio.Redirect
    yield rio.Session
    yield rio.TextStyle
    yield rio.Theme
    yield rio.UserSettings
    yield rio.DateChangeEvent

    for module in (rio.event, rio.fills):
        for name in module.__all__:
            yield getattr(module, name)

    # Yield all events. There is no perfectly safe way to detect these
    # automatically, but the name is a good hint.
    for name, obj in vars(rio).items():
        if not name.endswith("Event"):
            continue

        assert inspect.isclass(obj), obj
        yield obj

    # Yield classes that also need their children documented
    to_do = [rio.Component]

    while to_do:
        cur = to_do.pop()

        # Skip anything not in the `rio` module
        if not cur.__module__.startswith("rio."):
            continue

        # Queue the children
        to_do.extend(cur.__subclasses__())

        # Hardcoded items that DON'T need documentation
        if cur.__name__ in (
            # Built-in components
            "ClassContainer",
            "DialogContainer",
            "FundamentalComponent",
            # Components from examples
            "NavButton",
            "DefaultRootComponent",
        ):
            continue

        # Internal
        if cur.__name__.startswith("_"):
            continue

        yield cur


@functools.cache
def _get_unprocessed_docs() -> (
    t.Mapping[
        type | t.Callable,
        imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs,
    ]
):
    """
    Finds all documented objects and parses their docstrings, but doesn't
    post-process them. This is necessary because post-processing involves
    inserting links to other documented objects, which requires a full list of
    documented objects. So post-processing has to be done later in a separate
    step.
    """

    result: dict = {
        rio.URL: _make_docs_for_rio_url(),
    }

    # Use heuristics to find all objects which should likely be public
    for obj in _find_possibly_public_objects():
        # Parse the object's docs
        if isinstance(obj, type):
            docs = imy.docstrings.ClassDocs.from_class(obj)
        else:
            docs = imy.docstrings.FunctionDocs.from_function(obj)

        # Make the final determination whether this object is public
        if not docs.metadata.public:
            continue

        # Make the summary into a single line. (This is because the summary is
        # sometimes displayed inside a `rio.Text`, which honors newlines. We
        # don't want that.)
        if docs.summary:
            docs.summary = docs.summary.replace("\n", " ")

        # This object is public
        result[obj] = docs

    return result


@functools.cache
def find_documented_objects() -> (
    t.Mapping[
        type | t.Callable,
        imy.docstrings.ClassDocs | imy.docstrings.FunctionDocs,
    ]
):
    """
    Find all public classes and functions in `Rio` along with their
    documentation.
    """
    result = _get_unprocessed_docs()

    for docs in result.values():
        postprocess_docs(docs)

    return result


def postprocess_docs(
    docs: imy.docstrings.FunctionDocs | imy.docstrings.ClassDocs,
) -> None:
    """
    Applies rio-specific post-processing.
    """

    if isinstance(docs, imy.docstrings.FunctionDocs):
        postprocess_function_docs(docs)
    elif issubclass(docs.object, rio.Component):
        postprocess_component_docs(docs)
    else:
        postprocess_class_docs(docs)


def postprocess_function_docs(docs: imy.docstrings.FunctionDocs) -> None:
    if docs.summary is not None:
        docs.summary = insert_links_into_markdown(
            docs.summary, own_name=docs.name
        )

    if docs.details is not None:
        docs.details = insert_links_into_markdown(
            docs.details, own_name=docs.name
        )

    # Post-process the parameters
    for param_docs in docs.parameters:
        if param_docs.description:
            param_docs.description = insert_links_into_markdown(
                param_docs.description,
                own_name=f"{docs.name}.{param_docs.name}",
            )


def postprocess_class_docs(docs: imy.docstrings.ClassDocs) -> None:
    """
    Perform Rio specific post-processing on the component, such as stripping out
    internal attributes and functions.
    """

    # Strip out anything `Session` inherits from `unicall`
    if docs.name == "Session":
        to_remove = set(dir(unicall.Unicall)).difference(vars(rio.Session))
        docs.functions = [
            func for func in docs.functions if func.name not in to_remove
        ]

    # Strip default docstrings created by dataclasses
    if docs.summary is not None and docs.summary.startswith(f"{docs.name}("):
        docs.summary = None
        docs.details = None

    # Inserts links to other documentation
    if docs.summary is not None:
        docs.summary = insert_links_into_markdown(
            docs.summary, own_name=docs.name
        )

    if docs.details is not None:
        docs.details = insert_links_into_markdown(
            docs.details, own_name=docs.name
        )

    # Strip internal attributes
    index = 0
    while index < len(docs.attributes):
        prop = docs.attributes[index]

        # Decide whether to keep it
        keep = True

        keep = keep and not prop.name.startswith("_")
        keep = keep and prop.type is not dataclasses.KW_ONLY

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.attributes[index]

    # Additional per-attribute post-processing
    for prop in docs.attributes:
        # Insert links to other documentation pages
        if prop.description is not None:
            prop.description = insert_links_into_markdown(
                prop.description, own_name=f"{docs.name}.{prop.name}"
            )

    # Strip internal properties
    index = 0
    while index < len(docs.properties):
        prop = docs.properties[index]

        # Decide whether to keep it
        keep = not prop.name.startswith("_")

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.properties[index]

    # Additional per-property post-processing
    for prop in docs.properties:
        # Insert links to other documentation pages
        for func in (prop.getter, prop.setter):
            if func is None:
                continue

            if func.summary is not None:
                func.summary = insert_links_into_markdown(
                    func.summary, own_name=f"{docs.name}.{prop.name}"
                )

            if func.details is not None:
                func.details = insert_links_into_markdown(
                    func.details, own_name=f"{docs.name}.{prop.name}"
                )

    # Skip internal functions
    index = 0
    while index < len(docs.functions):
        func = docs.functions[index]

        # Decide whether to keep it

        # Internal methods start with an underscore
        keep = not func.name.startswith("_")

        # Some methods in components are meant to be used by the user, but only
        # when they're the one creating the component. For example, the `build`
        # method is only intended to be used by the component itself, and
        # documenting it would be pointless at best, and confusing at worst.
        is_inherited_protected_method = (
            docs.name != "Component"
            and func.name
            in (
                "build",
                "call_event_handler",
                "force_refresh",
            )
        )
        if is_inherited_protected_method:
            keep = False

        # Strip lambdas
        if func.name == "<lambda>":
            keep = False

        # Make sure to keep the constructor
        if func.name == "__init__":
            keep = True

        # Some classes are not meant to be constructed by the user. Strip their
        # constructor.
        if (
            docs.name
            in (
                "FileInfo",
                "Session",
            )
            and func.name == "__init__"
        ):
            keep = False

        # Check if it's explicitly excluded
        if not func.metadata.public:
            keep = False

        # Strip it out, if necessary
        if keep:
            index += 1
        else:
            del docs.functions[index]

    # Additional per-function post-processing
    for func_docs in docs.functions:
        # Strip the ridiculous default docstring created by dataclasses
        #
        # FIXME: Not working for some reason
        if (
            func_docs.name == "__init__"
            and func_docs.summary
            == "Initialize self. See help(type(self)) for accurate signature."
        ):
            func_docs.summary = None
            func_docs.details = None

        # Inject a short description for `__init__` if there is none.
        if func_docs.name == "__init__" and func_docs.summary is None:
            func_docs.summary = f"Creates a new `{docs.name}` instance."

        # Insert links to other documentation pages
        if func_docs.summary is not None:
            func_docs.summary = insert_links_into_markdown(
                func_docs.summary, own_name=f"{docs.name}.{func_docs.name}"
            )

        if func_docs.details is not None:
            func_docs.details = insert_links_into_markdown(
                func_docs.details, own_name=f"{docs.name}.{func_docs.name}"
            )

        # Post-process the parameters
        for param_docs in func_docs.parameters:
            if param_docs.description:
                param_docs.description = insert_links_into_markdown(
                    param_docs.description,
                    own_name=f"{docs.name}.{func_docs.name}.{param_docs.name}",
                )


def postprocess_component_docs(docs: imy.docstrings.ClassDocs) -> None:
    # Apply the standard class post-processing
    postprocess_class_docs(docs)

    # Remove the `bind()` method inherited from `rio.Component`, because
    # that method is only useful in custom components
    for i, func in enumerate(docs.functions):
        if func.name == "bind":
            del docs.functions[i]
            break

    # Subclasses of `rio.Component` inherit a load of constructor parameters,
    # which clutter the docs. We'll sort the keyword-only parameters so that the
    # inherited parameters appear at the end.
    if docs.object is not rio.Component:
        try:
            init_func = next(
                func for func in docs.functions if func.name == "__init__"
            )
        except StopIteration:
            pass
        else:
            parameters = list[imy.docstrings.FunctionParameter]()
            kwargs = list[imy.docstrings.FunctionParameter]()

            for param in init_func.parameters:
                if param.kw_only:
                    kwargs.append(param)
                else:
                    parameters.append(param)

            COMPONENT_CONSTRUCTOR_PARAMS = set(rio.Component.__annotations__)
            kwargs.sort(
                key=lambda param: param.name in COMPONENT_CONSTRUCTOR_PARAMS
            )
            parameters += kwargs

            init_func.parameters = parameters



__version__ = "0.10.5rc0"


# There is an issue with `rye test`. rye passes a `--rootdir` argument to
# pytest, and webview parses command line arguments when it is imported. It
# crashes parsing the `--rootdir` argument, so we'll temporarily remove the
# command line arguments while importing webview.
import sys

argv = sys.argv
sys.argv = argv[:1]

try:
    import webview  # noqa: F401
except ImportError:
    pass
finally:
    sys.argv = argv


import logging

_logger = logging.getLogger(__name__)

# Re-export dataclass stuff for easy use.
from dataclasses import KW_ONLY as KW_ONLY
from dataclasses import field as field

# URLs are used as an important datatype within rio. Re-export them for easy
# use.
from yarl import URL as URL

from . import event as event
from . import icons as icons  # For backwards compat. Delete eventually
from . import patches_for_3rd_party_stuff
from .app import *
from .color import *
from .components import *
from .cursor_style import *
from .dialog import *
from .errors import *
from .fills import *
from .routing import *
from .session import *
from .text_style import *
from .theme import *
from .user_settings_module import *
from .utils import *

del patches_for_3rd_party_stuff



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY, dataclass

from uniserde import JsonDoc

import rio
import rio.docs

from .. import utils
from .fundamental_component import FundamentalComponent

__all__ = [
    "FilePickEvent",
    "FilePickerArea",
]


@t.final
@rio.docs.mark_constructor_as_private
@dataclass
class FilePickEvent:
    """
    Holds information regarding a file upload event.

    This is a simple dataclass that stores useful information for when the user
    chooses a file using a `FilePickerArea`. You'll typically receive this as
    argument in `on_choose_file` events.

    ## Attributes

    `file`: Handle to the uploaded file.
    """

    file: rio.FileInfo


@t.final
class FilePickerArea(FundamentalComponent):
    """
    Drag & Drop are for files

    The `FilePickerArea` component allows the user to upload files either by
    dragging and dropping them onto the component, or optionally using a regular
    file browser. Whenever a file has been uploaded, the `on_file_upload` event
    is triggered, allowing you to run code.

    ## Attributes

    `content`: A custom text to display to the user. If not provided, the
        component will display a default message.

    `file_types`: A list of file extensions which the user is allowed
        to select. Defaults to `None`, which means that the user may select any
        file. Values can be passed as file extensions, ('pdf', '.pdf', '*.pdf'
        are all accepted) or MIME types (e.g. 'application/pdf').

        Note that there is no way for Rio to enforce the file type. Users will
        always be able to upload arbitrary files, e.g. by renaming them. Treat
        this as a hint to the user, and so the file browser may filter files,
        but not as a security measure.

    `on_choose_file`: Triggered when the user uploads a file, either by dragging
        and dropping it onto the component, or by selecting it in the file
        browser. The event handler receives a `FileInfo` object, which contains
        information about the uploaded file.


    ## Metadata

    `experimental`: True
    """

    _: KW_ONLY
    content: str | None = None
    file_types: list[str] | None = None
    on_choose_file: rio.EventHandler[FilePickEvent] = None

    def _custom_serialize_(self) -> JsonDoc:
        if self.file_types is None:
            return {}

        return {
            "file_types": list(
                {
                    utils.normalize_file_type(file_type)
                    for file_type in self.file_types
                }
            )
        }

    async def _on_file_upload_(self, files: list[rio.FileInfo]) -> None:
        for file in files:
            # TODO: Should these be called simultaneously?
            event_data = FilePickEvent(file)

            await self.call_event_handler(
                self.on_choose_file,
                event_data,
            )


FilePickerArea._unique_id_ = "FilePickerArea-builtin"



import typing as t
from pathlib import Path

import introspection.typing

from .. import components, global_state
from ..components.component import Component, ComponentMeta
from ..state_properties import (
    PleaseTurnThisIntoAnAttributeBinding,
    StateProperty,
)

__all__ = [
    "apply_monkeypatches",
]


def apply_monkeypatches() -> None:
    """
    Enables extra safeguards that are too slow to be enabled permanently. Used
    by `rio run` when running in debug mode.
    """
    introspection.wrap_method(Component_bind, Component, "bind")
    introspection.wrap_method(ComponentMeta_call, ComponentMeta, "__call__")
    introspection.wrap_method(StateProperty_get, StateProperty, "__get__")
    introspection.wrap_method(StateProperty_set, StateProperty, "__set__")
    introspection.wrap_method(LinearContainer_init, components.Row, "__init__")
    introspection.wrap_method(
        LinearContainer_init, components.Column, "__init__"
    )
    introspection.wrap_method(ListView_init, components.ListView, "__init__")
    introspection.wrap_method(
        AspectRatioContainer_init, components.AspectRatioContainer, "__init__"
    )


def Component_bind(wrapped_method, self: Component):
    if global_state.currently_building_session is None:
        raise RuntimeError(
            "`.bind()` can only be called inside of the `build()` method"
        )

    if global_state.currently_building_component is not self:
        raise RuntimeError("`.bind()` can only be called on `self`")

    return wrapped_method(self)


def ComponentMeta_call(wrapped_method, cls, *args, **kwargs):
    component: Component = wrapped_method(cls, *args, **kwargs)

    # Make sure the component was properly initialized
    try:
        component._properties_set_by_creator_
    except AttributeError:
        raise Exception(
            f"The component {component} wasn't initialized properly. If you"
            f" have written a custom `__init__` method, make sure to call"
            f" `super().__init__()`"
        ) from None

    # Keep track of who created this component
    with introspection.CallFrame.up(2) as frame:
        component._creator_stackframe_ = (
            Path(frame.file_name),
            frame.current_line_number,
        )

    # Track whether this instance is internal to Rio. This is the case if
    # this component's creator is defined in Rio.
    creator = global_state.currently_building_component
    if creator is None:
        assert type(component).__qualname__ == "HighLevelRootComponent", type(
            component
        ).__qualname__
        component._rio_internal_ = True
    else:
        component._rio_internal_ = creator._rio_builtin_

    return component


def StateProperty_get(
    wrapped_method,
    self: StateProperty,
    instance: Component | None,
    owner=None,
):
    if instance is None:
        return self

    if not instance._init_called_:
        raise RuntimeError(
            f"The `__init__` method of {instance} attempted to access the state"
            f" property {self.name!r}. This is not allowed. `__init__` methods"
            f" must only *set* properties, not *read* them. Move the code that"
            f" needs to read a state property into the `__post_init__` method."
        )

    return wrapped_method(self, instance, owner)


def StateProperty_set(
    wrapped_method,
    self: StateProperty,
    instance: Component,
    value: object,
) -> None:
    # Type check the value
    if not isinstance(value, PleaseTurnThisIntoAnAttributeBinding):
        try:
            annotation = self._resolved_annotation
        except AttributeError:
            annotation = introspection.typing.resolve_forward_refs(
                self._raw_annotation,
                self._module,
                mode="ast",
                strict=False,
                treat_name_errors_as_imports=True,
            )
            self._resolved_annotation = annotation

        try:
            valid = introspection.typing.is_instance(
                value, annotation, forward_ref_context=self._module
            )
        except introspection.errors.CannotResolveForwardref as error:
            # revel.warning(
            #     f"Unable to verify assignment of {value!r} to"
            #     f" {type(instance).__qualname__}.{self.name} (annotated as"
            #     f" {self._annotation_as_string()}) because the forward"
            #     f" reference {error.forward_ref!r} cannot be resolved",
            #     markup=False,
            # )
            pass
        except NotImplementedError:
            # revel.warning(
            #     f"Unable to verify assignment of {value!r} to"
            #     f" {type(instance).__qualname__}.{self.name} (annotated as"
            #     f" {self._annotation_as_string()})",
            #     markup=False,
            # )
            pass
        else:
            if not valid:
                raise TypeError(
                    f"The value {value!r} can't be assigned to"
                    f" {type(instance).__qualname__}.{self.name}, which is"
                    f" annotated as {self._annotation_as_string()}"
                )

    # Chain to the original method
    wrapped_method(self, instance, value)


def LinearContainer_init(
    wrapped_method,
    self: components.Row,
    *children,
    proportions: t.Literal["homogeneous"] | t.Sequence[float] | None = None,
    **kwargs,
) -> None:
    # Proportions related checks
    if proportions is not None and not isinstance(proportions, str):
        proportions = list(proportions)

        # Make sure the number of proportions matches the number of children
        if len(proportions) != len(children):
            if len(proportions) == 1:
                proportions_str = f"one proportion was {proportions}"
            else:
                proportions_str = f"{len(proportions)} proportions were"

            raise ValueError(
                f"The component has {len(children)} children, but {proportions_str} provided."
            )

        # The sum of all proportions must exceed 0
        if proportions and sum(proportions) <= 0:
            # TODO: While this handles a length of zero children fine, JS
            # probably doesn't!
            raise ValueError(
                "The sum of all proportions must be greater than 0."
            )

        # Every proportion must be positive
        for proportion in proportions:
            if proportion < 0:
                raise ValueError(
                    f"{proportion} is not a valid proportion. All proportions must be greater than or equal to zero."
                )

    # Chain to the original method
    wrapped_method(
        self,
        *children,
        proportions=proportions,
        **kwargs,
    )


def ListView_init(
    wrapped_method,
    self: components.ListView,
    *children,
    **kwargs,
) -> None:
    # Make sure all children have a key set
    assert isinstance(children, tuple), children

    for child in children:
        if child.key != None or isinstance(child, components.SeparatorListItem):
            continue

        raise ValueError(
            f"ListView child {child!r} has no key set. List items change frequently, and are often reshuffled. This can lead to unexpected reconciliations, and slow down the UI."
        )

    # Chain to the original method
    wrapped_method(self, *children, **kwargs)


def AspectRatioContainer_init(
    wrapped_method,
    self: components.AspectRatioContainer,
    content: components.Component,
    aspect_ratio: float,
    *args,
    **kwargs,
) -> None:
    # Make sure the aspect ratio is valid
    if aspect_ratio <= 0:
        raise ValueError(
            f"The aspect ratio must be greater than zero, but was {aspect_ratio}"
        )

    # Chain to the original method
    return wrapped_method(self, content, aspect_ratio, *args, **kwargs)



import enum

__all__ = ["CursorStyle"]


class CursorStyle(enum.Enum):
    """
    Enumeration of all available cursor styles. Use these to indicate which
    kinds of action a user can perform while hovering over a component.

    ## Example

    ```python
    rio.Rectangle(
        fill=rio.Color.WHITE,
        cursor=rio.CursorStyle.POINTER,
    )
    ```
    """

    DEFAULT = enum.auto()
    NONE = enum.auto()
    HELP = enum.auto()
    POINTER = enum.auto()
    LOADING = enum.auto()  # "wait" in CSS
    BACKGROUND_LOADING = enum.auto()  # "progress" in CSS
    CROSSHAIR = enum.auto()
    TEXT = enum.auto()
    MOVE = enum.auto()
    NOT_ALLOWED = enum.auto()
    CAN_GRAB = enum.auto()  # "grab" in CSS
    IS_GRABBED = enum.auto()  # "grabbing" in CSS
    ZOOM_IN = enum.auto()
    ZOOM_OUT = enum.auto()



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY, dataclass

import rio.docs

from .component import Component

__all__ = [
    "NumberInput",
    "NumberInputChangeEvent",
    "NumberInputConfirmEvent",
    "NumberInputFocusEvent",
]


# These must be ints so that `integer * multiplier` returns an int and not a
# float
_multiplier_suffixes: t.Mapping[str, int] = {
    "k": 1_000,
    "m": 1_000_000,
}


@t.final
@rio.docs.mark_constructor_as_private
@dataclass
class NumberInputChangeEvent:
    """
    Holds information regarding a number change event.

    This is a simple dataclass that stores useful information for when the user
    changes a number. You'll typically receive this as argument in `on_change`
    events.

    ## Attributes

    `value`: The new `value` of the `NumberInput`.
    """

    value: float


@t.final
@rio.docs.mark_constructor_as_private
@dataclass
class NumberInputConfirmEvent:
    """
    Holds information regarding a number confirm event.

    This is a simple dataclass that stores useful information for when the user
    confirms a number. You'll typically receive this as argument in `on_confirm`
    events.

    ## Attributes

    `value`: The new `value` of the `NumberInput`.
    """

    value: float


@t.final
@rio.docs.mark_constructor_as_private
@dataclass
class NumberInputFocusEvent:
    """
    Holds information regarding a number input focus event.

    This is a simple dataclass that stores useful information for when a
    `NumberInput` gains or loses focus. You'll typically receive this as
    argument in `on_gain_focus` and `on_lose_focus` events.

    ## Attributes

    value: The `value` of the `NumberInput`.
    """

    value: float


@t.final
class NumberInput(Component):
    """
    Like `TextInput`, but specifically for inputting numbers.

    `NumberInput` allows the user to enter a number. This is similar to the
    `TextInput` component, but with some goodies specifically for handling
    numbers. The value is automatically parsed and formatted according to the
    user's locale, and you can specify minimum and maximum values to limit the
    user's input.

    Number inputs go way beyond just parsing the input using `int` or `float`.
    They try their best to understand what the user is trying to enter, and also
    support suffixes such as "k" and "m" to represent thousands and millions
    respectively.


    ## Attributes

    `value`: The number currently entered by the user.

    `label`: A short text to display next to the number input.

    `style`: Changes the visual appearance of the text input.

    `prefix_text`: A short text to display before the number input. Useful for
        displaying currency symbols or other prefixed units.

    `suffix_text`: A short text to display after the number input. Useful for
        displaying currency names or units.

    `minimum`: The minimum value the number can be set to.

    `maximum`: The maximum value the number can be set to.

    `decimals`: The number of decimals to accept. If the user enters more
        decimals, they will be rounded off. If this value is equal to `0`, the
        input's `value` is guaranteed to be an integer, rather than float.

    `thousands_separator`: By default, the number is formatted according to the
        user's locale. This means numbers will show up correctly for everyone,
        regardless of where they live and which thousands separator they use.
        If you want to override this behavior, you can set this attribute to
        `False`. This will disable the thousands separator altogether.
        Alternatively, provide a custom string to use as the thousands
        separator.

    `is_sensitive`: Whether the text input should respond to user input.

    `is_valid`: Visually displays to the user whether the current text is
        valid. You can use this to signal to the user that their input needs to
        be changed.

    `on_change`: Triggered when the user changes the number.

    `on_confirm`: Triggered when the user explicitly confirms their input,
        such as by pressing the "Enter" key.


    ## Examples

    Here's a simple example that allows the user to select a value and displays
    it back to them:

    ```python
    class MyComponent(rio.Component):
        value: float = 1

        def build(self) -> rio.Component:
            return rio.Column(
                rio.NumberInput(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the text.
                    value=self.bind().value,
                    minimum=1,
                    maximum=10,
                ),
                rio.Text(f"You've selected: {self.value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    changes the text:

    ```python
    class MyComponent(rio.Component):
        value: float = 1

        def on_value_change(self, event: rio.NumberInputChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.value = event.value
            print(f"You've selected: {self.value}")

        def build(self) -> rio.Component:
            return rio.NumberInput(
                value=self.value,
                minimum=1,
                maximum=10,
                on_change=self.on_value_change,
            )
    ```
    """

    value: float = 0
    _: KW_ONLY
    label: str = ""
    accessibility_label: str = ""
    style: t.Literal["underlined", "rounded", "pill"] = "underlined"
    prefix_text: str = ""
    suffix_text: str = ""
    minimum: float | None = None
    maximum: float | None = None
    decimals: int = 2
    thousands_separator: bool | str = True
    is_sensitive: bool = True
    is_valid: bool = True

    on_change: rio.EventHandler[NumberInputChangeEvent] = None
    on_confirm: rio.EventHandler[NumberInputConfirmEvent] = None

    on_gain_focus: rio.EventHandler[NumberInputFocusEvent] = None
    on_lose_focus: rio.EventHandler[NumberInputFocusEvent] = None

    def __post_init__(self):
        self._text_input = None

    def _try_set_value(self, raw_value: str) -> bool:
        """
        Parse the given string and update the component's value accordingly.
        Returns `True` if the value was successfully updated, `False` otherwise.
        """

        # Strip the number down as much as possible
        raw_value = raw_value.strip()
        raw_value = raw_value.replace(self.session._thousands_separator, "")

        # If left empty, set the value to 0, if that's allowable
        if not raw_value:
            self.value = 0

            if self.minimum is not None and self.minimum > 0:
                self.value = self.minimum

            if self.maximum is not None and self.maximum < 0:
                self.value = self.maximum

            return True

        # Check for a multiplier suffix
        suffix = raw_value[-1].lower()
        multiplier = 1

        if suffix.isalpha():
            try:
                multiplier = _multiplier_suffixes[suffix]
            except KeyError:
                pass
            else:
                raw_value = raw_value[:-1].rstrip()

        # Try to parse the number
        try:
            value = float(
                raw_value.replace(self.session._decimal_separator, ".")
            )
        except ValueError:
            self.value = self.value  # Force the old value to stay
            return False

        # Apply the multiplier
        value *= multiplier

        # Limit the number of decimals
        #
        # Ensure the value is an integer, if the number of decimals is 0
        value = round(value, None if self.decimals == 0 else self.decimals)

        # Clamp the value
        minimum = self.minimum
        if minimum is not None:
            value = max(value, minimum)

        maximum = self.maximum
        if maximum is not None:
            value = min(value, maximum)

        # Update the value
        self.value = value
        return True

    async def _on_gain_focus(self, ev: rio.TextInputFocusEvent) -> None:
        await self.call_event_handler(
            self.on_gain_focus,
            NumberInputFocusEvent(self.value),
        )

    async def _on_lose_focus(self, ev: rio.TextInputFocusEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_change,
                NumberInputChangeEvent(self.value),
            )

        await self.call_event_handler(
            self.on_lose_focus,
            NumberInputFocusEvent(self.value),
        )

    async def _on_confirm(self, ev: rio.TextInputConfirmEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_change,
                NumberInputChangeEvent(self.value),
            )

            await self.call_event_handler(
                self.on_confirm,
                NumberInputConfirmEvent(self.value),
            )

    def _formatted_value(self) -> str:
        """
        Convert the given number to a string, formatted according to the
        component's settings.
        """
        # Otherwise use the locale's settings
        if self.decimals == 0:
            int_str, frac_str = f"{self.value:.0f}", ""
        else:
            int_str, frac_str = f"{self.value:.{self.decimals}f}".split(".")

        # Add the thousands separators
        if self.thousands_separator is True:
            thousands_separator = self._session_._thousands_separator
        elif self.thousands_separator is False:
            thousands_separator = ""
        else:
            thousands_separator = self.thousands_separator

        integer_part_with_sep = f"{int(int_str):,}".replace(
            ",", thousands_separator
        )

        # Construct the final formatted number
        if self.decimals == 0:
            return integer_part_with_sep

        return f"{integer_part_with_sep}{self.session._decimal_separator}{frac_str}"

    def build(self) -> rio.Component:
        # Build the component
        self._text_input = rio.TextInput(
            text=self._formatted_value(),
            label=self.label,
            style=self.style,
            prefix_text=self.prefix_text,
            suffix_text=self.suffix_text,
            is_sensitive=self.is_sensitive,
            is_valid=self.is_valid,
            accessibility_label=self.accessibility_label,
            on_confirm=self._on_confirm,
            on_gain_focus=self._on_gain_focus,
            on_lose_focus=self._on_lose_focus,
        )
        return self._text_input

    async def grab_keyboard_focus(self) -> None:
        """
        ## Metadata

        `public`: False
        """
        if self._text_input is not None:
            await self._text_input.grab_keyboard_focus()



import aiofiles
from pytest import MonkeyPatch
from uniserde import JsonDoc

import rio.testing


class FakeFile:
    def __init__(self, content: str):
        self._content = content

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass

    async def read(self) -> str:
        return self._content


class RootSettings(rio.UserSettings):
    foo: str = "foo"


class FooSettings(rio.UserSettings):
    section_name = "foo"

    bar: str = "bar"


async def test_load_settings():
    user_settings: JsonDoc = {
        ":foo": "bar",
        "foo:bar": "baz",
    }

    async with rio.testing.TestClient(
        running_in_window=False,
        default_attachments=(RootSettings(), FooSettings()),
        user_settings=user_settings,
    ) as test_client:
        root_settings = test_client.session[RootSettings]
        foo_settings = test_client.session[FooSettings]

        assert root_settings.foo == "bar"
        assert foo_settings.bar == "baz"


async def test_load_settings_file(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(
        aiofiles,
        "open",
        lambda _: FakeFile('{"foo": "bar", "section:foo": {"bar": "baz"} }'),
    )

    async with rio.testing.TestClient(
        running_in_window=True,
        default_attachments=(RootSettings(), FooSettings()),
    ) as test_client:
        root_settings = test_client.session[RootSettings]
        foo_settings = test_client.session[FooSettings]

        assert root_settings.foo == "bar"
        assert foo_settings.bar == "baz"



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Separator",
]


@t.final
class Separator(FundamentalComponent):
    """
    A line to separate content.

    `Separator` creates a horizontal or vertical line that can be used to
    separate content. You can optionally assign a color.

    It is generally preferable to separate components using whitespace over
    explicit separators. Whitespace tends to look cleaner, while separators can
    add clutter. Separators can still absolutely make sense though, so here they
    are!


    ## Attributes

    `color`: The color of the `Separator`. If `None`, the color will be
        determined by the theme.


    ## Examples

    This example will display a two text components separated by a line:

    ```python
    rio.Row(
        rio.Text("First"),
        rio.Separator(),
        rio.Text("Second"),
        spacing=0.5,
    )
    ```
    """

    _: KW_ONLY
    color: rio.Color | None = None


Separator._unique_id_ = "Separator-builtin"



"""
This module exposes functionality which may or may not exist, depending on which
modules are available on the system.

Some values are not initialized at all, if the required module isn't in
`sys.modules`. This avoids long startup times if modules are available but not
used.
"""

from __future__ import annotations

import sys
import typing as t

import introspection

if t.TYPE_CHECKING:
    import matplotlib.axes  # type: ignore
    import matplotlib.figure  # type: ignore
    import numpy  # type: ignore
    import pandas  # type: ignore
    import plotly.graph_objects  # type: ignore
    import polars  # type: ignore

_IS_INITIALIZED = False

T = t.TypeVar("T")


FLOAT_TYPES = ()
INT_TYPES = ()
BOOL_TYPES = ()
STR_TYPES = ()

NUMPY_ARRAY_TYPES: tuple[type[numpy.ndarray], ...] = ()

PANDAS_DATAFRAME_TYPES: tuple[type[pandas.DataFrame], ...] = ()
POLARS_DATAFRAME_TYPES: tuple[type[polars.DataFrame], ...] = ()

PLOTLY_GRAPH_TYPES: tuple[type[plotly.graph_objects.Figure], ...] = ()
MATPLOTLIB_GRAPH_TYPES: tuple[type, ...] = ()
MATPLOTLIB_AXES_TYPES: tuple[type[matplotlib.axes.Axes], ...] = ()

# This is a mapping of "weird" types to the "canonical" type, like `{np.int8: int}`
TYPE_NORMALIZERS: dict[type[T], t.Callable[[T], T]] = {}  # type: ignore


def initialize(force: bool = False) -> None:
    """
    If called for the first time, initialize all constants in the module. This
    is not automatically done on module load, to make sure any needed modules
    have already been imported - some functionality is not initialized if those
    other modules aren't used.
    """
    global _IS_INITIALIZED
    global FLOAT_TYPES, INT_TYPES, BOOL_TYPES, STR_TYPES
    global NUMPY_ARRAY_TYPES
    global PANDAS_DATAFRAME_TYPES, POLARS_DATAFRAME_TYPES
    global PLOTLY_GRAPH_TYPES, MATPLOTLIB_GRAPH_TYPES

    # Already initialized?
    if _IS_INITIALIZED and not force:
        return

    _IS_INITIALIZED = True

    FLOAT_TYPES = (float, int)
    INT_TYPES = (int,)
    BOOL_TYPES = (bool,)
    STR_TYPES = (str,)

    # Is numpy available and loaded?
    if "numpy" in sys.modules:
        import numpy  # type: ignore

        NUMPY_ARRAY_TYPES = (numpy.ndarray,)

        numpy_floats = tuple(introspection.iter_subclasses(numpy.floating))
        numpy_ints = tuple(introspection.iter_subclasses(numpy.integer))
        numpy_bools = tuple(introspection.iter_subclasses(numpy.bool_))
        numpy_strings = tuple(introspection.iter_subclasses(numpy.str_))

        FLOAT_TYPES = (*FLOAT_TYPES, *numpy_floats, *numpy_ints)
        INT_TYPES += numpy_ints
        BOOL_TYPES += numpy_bools
        STR_TYPES += numpy_strings

    if "pandas" in sys.modules:
        import pandas  # type: ignore

        PANDAS_DATAFRAME_TYPES = (pandas.DataFrame,)

    if "polars" in sys.modules:
        import polars  # type: ignore

        POLARS_DATAFRAME_TYPES = (polars.DataFrame,)

    if "plotly" in sys.modules:
        import plotly.graph_objects  # type: ignore

        PLOTLY_GRAPH_TYPES = (plotly.graph_objects.Figure,)

    if "matplotlib" in sys.modules:
        import matplotlib.axes  # type: ignore
        import matplotlib.figure  # type: ignore

        MATPLOTLIB_AXES_TYPES = (matplotlib.axes.Axes,)
        MATPLOTLIB_GRAPH_TYPES = (
            matplotlib.figure.Figure,
        ) + MATPLOTLIB_AXES_TYPES

    # Populate our mapping of type normalizers
    for canonical_type, weird_types in (
        (str, STR_TYPES),
        (bool, BOOL_TYPES),
        (int, INT_TYPES),
        (float, FLOAT_TYPES),
    ):
        for weird_type in weird_types:
            TYPE_NORMALIZERS[weird_type] = canonical_type  # type: ignore



"""
Pretty-strings a traceback. The result looks very similar to Python's default,
but is colored and just tweaked in general.
"""

import html
import io
import linecache
import sys
import traceback
import typing as t
from pathlib import Path

import revel


def _format_single_exception_raw(
    out: t.IO[str],
    err: BaseException,
    *,
    include_header: bool,
    escape: t.Callable[[str], str],
    bold: str,
    nobold: str,
    dim: str,
    nodim: str,
    yellow: str,
    noyellow: str,
    red: str,
    nored: str,
    relpath: Path | None = None,
    frame_filter: t.Callable[[traceback.FrameSummary], bool],
) -> None:
    # Get the traceback as a list of frames
    tb_list = traceback.extract_tb(err.__traceback__)

    # Syntax errors are very special snowflakes and need separate treatment.
    #
    # The exact arguments depend on the python version.
    #
    # TODO: Is there a better way to do this? Since Python obviously isn't
    # keeping the arguments consistent, this could potentially break every
    # single Python version.
    if isinstance(err, SyntaxError):
        if sys.version_info < (3, 11):
            frame = traceback.FrameSummary(
                filename=err.filename,  # type: ignore
                lineno=err.lineno,
                name="<module>",
                line=err.text,
                locals=None,
            )
        else:
            frame = traceback.FrameSummary(
                filename=err.filename,  # type: ignore
                lineno=err.lineno,
                end_lineno=err.end_lineno,
                colno=err.offset,
                end_colno=err.end_offset,
                name="<module>",
                line=err.text,
                locals=None,
            )

        tb_list.append(frame)

    # Lead-in
    if include_header:
        out.write(f"{dim}Traceback (most recent call last):{nodim}\n")

    # Iterate through frames
    for frame in tb_list:
        # Keep this frame?
        if not frame_filter(frame):
            continue

        # Make paths relative to the relpath if they're inside it
        frame_path = Path(frame.filename)

        if relpath is not None and frame_path.is_absolute():
            try:
                frame_path = frame_path.relative_to(relpath)
            except ValueError:
                pass

        # Format the file location
        out.write(
            f"  {dim}File{nodim} {yellow}{escape(str(frame_path))}{noyellow}{dim}, {nodim}{yellow}line {frame.lineno}{noyellow}{dim}, in {escape(frame.name)}{nodim}\n"
        )

        # Display the source code from that line
        #
        # Insanely, the line in the frame has been stripped. Fetch it again.
        assert frame.lineno is not None  # Doesn't seem to happen, ever
        line = linecache.getline(frame.filename, frame.lineno)

        if line.strip():
            # If this is the last line, highlight the error
            if (
                frame is tb_list[-1]
                and hasattr(frame, "colno")
                and hasattr(frame, "end_colno")
            ):
                assert frame.colno is not None  # type: ignore
                assert frame.end_colno is not None  # type: ignore

                before = escape(line[: frame.colno])  # type: ignore
                error = escape(line[frame.colno : frame.end_colno])  # type: ignore
                after = escape(line[frame.end_colno :])  # type: ignore
                line = f"{before}{red}{error}{nored}{after}"
            else:
                line = escape(line)

            # NOW strip it
            line = line.strip()
            out.write(f"    {line}\n")

    # Actual error message
    out.write("\n")
    out.write(f"{bold}{red}{type(err).__name__}{nored}{nobold}")

    error_message = str(err)
    if error_message:
        out.write(f"{bold}: {escape(str(err))}{nobold}")


def format_exception_raw(
    err: BaseException,
    *,
    escape: t.Callable[[str], str],
    bold: str,
    nobold: str,
    dim: str,
    nodim: str,
    yellow: str,
    noyellow: str,
    red: str,
    nored: str,
    relpath: Path | None = None,
    frame_filter: t.Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    def format_inner(err: BaseException) -> None:
        # Chain to the cause or context if there is one
        if err.__cause__ is not None:
            format_inner(err.__cause__)
            out.write("\n\n")
            out.write(
                f"The above exception was the direct cause of the following:\n\n"
            )
            include_header = False

        elif err.__context__ is not None:
            format_inner(err.__context__)
            out.write("\n\n")
            out.write(
                f"During handling of the above exception, another exception occurred:\n\n"
            )
            include_header = False
        else:
            include_header = True

        # Format this exception
        _format_single_exception_raw(
            out,
            err,
            include_header=include_header,
            escape=escape,
            bold=bold,
            nobold=nobold,
            dim=dim,
            nodim=nodim,
            yellow=yellow,
            noyellow=noyellow,
            red=red,
            nored=nored,
            relpath=relpath,
            frame_filter=frame_filter,
        )

    # Start the recursion
    out = io.StringIO()
    format_inner(err)

    # Return the result
    return out.getvalue()


def format_exception_revel(
    err: BaseException,
    *,
    relpath: Path | None = None,
    frame_filter: t.Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    return format_exception_raw(
        err,
        bold="[bold]",
        nobold="[/]",
        dim="[dim]",
        nodim="[/]",
        yellow="[yellow]",
        noyellow="[/]",
        red="[red]",
        nored="[/]",
        escape=revel.escape,
        relpath=relpath,
        frame_filter=frame_filter,
    )


def format_exception_html(
    err: BaseException,
    *,
    relpath: Path | None = None,
    frame_filter: t.Callable[[traceback.FrameSummary], bool] = lambda _: True,
) -> str:
    result = format_exception_raw(
        err,
        bold='<span class="rio-traceback-bold">',
        nobold="</span>",
        dim='<span class="rio-traceback-dim">',
        nodim="</span>",
        yellow='<span class="rio-traceback-yellow">',
        noyellow="</span>",
        red='<span class="rio-traceback-red">',
        nored="</span>",
        escape=html.escape,
        relpath=relpath,
        frame_filter=frame_filter,
    )

    return result.replace("\n", "<br>")



import subprocess
import sys
from pathlib import Path

import rio.utils

PROJECT_ROOT_DIR = rio.utils.PROJECT_ROOT_DIR
INPUT_DIR = PROJECT_ROOT_DIR / "frontend"
OUTPUT_DIR = PROJECT_ROOT_DIR / "rio" / "frontend files"
ASSETS_DIR = OUTPUT_DIR / "assets"


def build(*extra_args: str) -> None:
    # Build with vite
    npx(
        "vite",
        "build",
        INPUT_DIR,
        "--outDir",
        OUTPUT_DIR,
        "--config",
        PROJECT_ROOT_DIR / "vite.config.mjs",
        # The real base URL is only known at runtime, and may differ for each
        # session. This string here is an intentionally easy-to-replace
        # placeholder.
        "--base",
        "/rio-base-url-placeholder/rio/frontend/",
        "--emptyOutDir",
        *extra_args,
    )


def dev_build() -> None:
    build("--mode", "development", "--minify", "false")


def npx(*args: str | Path) -> None:
    subprocess.run(
        ["npx", *map(str, args)],
        check=True,
        shell=sys.platform == "win32",
    )



from __future__ import annotations

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "AspectRatioContainer",
]


class AspectRatioContainer(FundamentalComponent):
    """
    Contains a single child, maintaining a fixed aspect ratio.

    `AspectRatioContainer` is a container which contains just a single child.
    What makes it special, is that it always maintains a fixed aspect ratio,
    i.e. width to height ratio. This is useful if you've got a component that
    you want to be able to resize, without distorting it.


    ## Attributes

    `content`: The component to place inside the container.

    `aspect_ratio`: The aspect ratio to maintain. This is the desired width
        divided by the desired height.


    ## Examples

    This will create a Rectangle that scales up to fit the parent container, but
    always maintains a 16:9 aspect ratio:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.AspectRatioContainer(
                content=rio.Rectangle(
                    content=rio.Text(
                        "Hello World!",
                        justify="center",
                    ),
                    fill=rio.Color.RED,
                ),
                aspect_ratio=16 / 9,
            )
    ```


    ## Metadata

    `public`: False

    `experimental`: True
    """

    content: rio.Component
    aspect_ratio: float


AspectRatioContainer._unique_id_ = "AspectRatioContainer-builtin"



from __future__ import annotations

import typing as t

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "CodeExplorer",
]


class CodeExplorer(FundamentalComponent):
    """
    ## Metadata

    `public`: False
    """

    source_code: str
    build_result: rio.Component
    line_indices_to_component_keys: t.Sequence[str | int | None]

    style: t.Literal["horizontal", "vertical"] = "horizontal"

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "line_indices_to_component_keys": self.line_indices_to_component_keys,  # type: ignore
        }


CodeExplorer._unique_id_ = "CodeExplorer-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY

from uniserde import JsonDoc

import rio

from ...components.fundamental_component import FundamentalComponent

__all__ = ["ComponentTree"]


class ComponentTree(FundamentalComponent):
    """
    Note: This component makes no attempt to request the correct amount of
    space. Specify a width/height manually, or make sure it's in a properly
    sized parent.

    ## Metadata

    `public`: False
    """

    component_id: int  # This can be invalid. The component must deal with it.

    _: KW_ONLY

    # Triggered whenever the user selects a component in the tree. The passed
    # value is the component's ID.
    on_select_component: rio.EventHandler[int] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"component_id"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger the change event
        try:
            new_id = delta_state["component_id"]
        except KeyError:
            pass
        else:
            assert isinstance(new_id, int), new_id
            await self.call_event_handler(self.on_select_component, new_id)


ComponentTree._unique_id_ = "ComponentTree-builtin"



from ...components.fundamental_component import FundamentalComponent

__all__ = ["ComponentPicker"]


class ComponentPicker(FundamentalComponent):
    """
    Lets the user select a component in the ComponentTree by clicking on it in
    the DOM.

    ## Metadata

    `public`: False
    """


ComponentPicker._unique_id_ = "ComponentPicker-builtin"



from __future__ import annotations

import typing as t
from dataclasses import KW_ONLY, dataclass

from uniserde import JsonDoc

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Switch",
    "SwitchChangeEvent",
]


@t.final
@dataclass
class SwitchChangeEvent:
    """
    Holds information regarding a switch change event.

    This is a simple dataclass that stores useful information for when the user
    switches a `Switch` on or off. You'll typically receive this as argument in
    `on_change` events.

    ## Attributes

    `is_on`: Whether the switch is now turned on.
    """

    is_on: bool


@t.final
class Switch(FundamentalComponent):
    """
    An input for `True` / `False` values.

    Switches allow the user to toggle between an "on" and an "off" state and
    effectively correspond to a Python `bool` value. Use them to allow the user
    to enable or disable certain features, or to select between two options.

    These are very similar to the `Checkbox` component, but their visuals can
    sometimes be preferable depending on the context.


    ## Attributes

    `is_on`: Whether the switch is currently in the "on" state.

    `is_sensitive`: Whether the switch should respond to user input.

    `on_change`: Triggered when the user toggles the switch.


    ## Examples

    Here's a simple example that allows the user to turn a switch on and off and
    displays the current state:

    ```python
    class MyComponent(rio.Component):
        is_on: bool = False

        def build(self) -> rio.Component:
            return rio.Column(
                rio.Switch(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the text.
                    is_on=self.bind().is_on,
                ),
                rio.Text("On" if self.is_on else "Off"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    changes the text:

    ```python
    class MyComponent(rio.Component):
        is_on: bool = False

        def on_value_change(self, event: rio.SwitchChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.is_on = event.is_on
            print("The switch is now", "On" if self.is_on else "Off")

        def build(self) -> rio.Component:
            return rio.Switch(
                is_on=self.is_on,
                on_change=self.on_value_change,
            )
    ```
    """

    is_on: bool = False
    _: KW_ONLY
    is_sensitive: bool = True
    on_change: rio.EventHandler[SwitchChangeEvent] = None

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"is_on"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

        if "is_on" in delta_state and not self.is_sensitive:
            raise AssertionError(
                f"Frontend tried to set `Switch.is_on` even though `is_sensitive` is `False`"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_on"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                SwitchChangeEvent(new_value),
            )


Switch._unique_id_ = "Switch-builtin"



import typing as t
from dataclasses import KW_ONLY

from uniserde import JsonDoc

from .. import deprecations
from .fundamental_component import FundamentalComponent

__all__ = [
    "Markdown",
]


@t.final
class Markdown(FundamentalComponent):
    '''
    Displays Markdown-formatted text.

    Markdown is a lightweight markup language that allows you to write text with
    simple formatting, such as **bold**, *italics*, and links. The `Markdown`
    component displays text formatted with the Markdown syntax.

    Markdown is a great way to write text that is both human-readable, yet
    beautifully formatted.


    ## Attributes

    `text`: The markdown-formatted text to display.

    `default_language`: The default language to use for code blocks. If
        `None`, Rio will try to guess the language automatically. If a
        default is given, it will be used for all code blocks that don't
        specify a language explicitly.

        Inline code will always use the default language, since they are too
        short to reliably guess the language - so make sure to set a default
        language if you want your inline code to be syntax-highlighted.

    `selectable`: Whether the text can be selected by the user.

    `justify`: Controls the alignment of the text within each individual line.

    `overflow`: What to do with long text. If this is `"nowrap"`, the text will
        be as wide as it needs to be to fit without wrapping. If `"wrap"`, the
        text will wrap to the next line when running out of horizontal space.
        Finally, if `"ellipsize"`, the text will be truncated when there isn't
        enough space and an ellipsis (`...`) will be added.


    ## Examples

    This example will display a short markdown-formatted text:

    ```python
    rio.Markdown(
        """
    # Hello, world!

    I am a **Markdown** component and my job is to display _formatted_ text.
    """
    )
    ```
    '''

    text: str
    _: KW_ONLY
    default_language: str | None = None
    selectable: bool = True
    justify: t.Literal["left", "right", "center", "justify"] = "left"
    wrap: bool | t.Literal["ellipsize"] = True
    overflow: t.Literal["nowrap", "wrap", "ellipsize"] = "wrap"

    def _custom_serialize_(self) -> JsonDoc:
        # The old `wrap` attribute has been replaced with `overflow`. Remap the
        # value.
        if self.wrap is not True:
            deprecations.warn_parameter_renamed(
                since="0.10",
                old_name="wrap",
                new_name="overflow",
                owner="rio.Markdown",
            )

        if self.wrap is False:
            overflow = "nowrap"
        elif self.wrap == "ellipsize":
            overflow = "ellipsize"
        else:
            overflow = self.overflow

        # Build the result
        return {
            "overflow": overflow,
        }


Markdown._unique_id_ = "Markdown-builtin"
